from __future__ import annotations

from typing import TYPE_CHECKING

from ._feature import Feature

if TYPE_CHECKING:
    from collections.abc import Sequence


def Time() -> Feature:
    """A feature to hold the integer time frame a node is in.

    Returns:
        Feature: A feature dict representing time
    """
    return {
        "feature_type": "node",
        "value_type": "int",
        "num_values": 1,
        "display_name": "Time",
        "required": True,
        "default_value": None,
    }


def Position(axes: Sequence[str]) -> Feature:
    """A feature to hold the position of a node (time not included).

    Args:
        axes (Sequence[str]): A sequence of the axis names. Used to infer the number
            of dimensions of the position values.

    Returns:
        Feature: A feature dict representing position
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": len(axes),
        "display_name": axes,
        "required": True,
        "default_value": None,
    }



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import networkx as nx
import numpy as np

from funtracks.features import FeatureDict

from .tracks import Tracks

if TYPE_CHECKING:
    from pathlib import Path

    from funtracks.annotators import TrackAnnotator

    from .tracks import Node


class SolutionTracks(Tracks):
    """Difference from Tracks: every node must have a track_id"""

    def __init__(
        self,
        graph: nx.DiGraph,
        segmentation: np.ndarray | None = None,
        time_attr: str | None = None,
        pos_attr: str | tuple[str] | list[str] | None = None,
        tracklet_attr: str | None = None,
        scale: list[float] | None = None,
        ndim: int | None = None,
        features: FeatureDict | None = None,
    ):
        """Initialize a SolutionTracks object.

        SolutionTracks extends Tracks to ensure every node has a track_id. A
        TrackAnnotator is automatically added to manage track IDs.

        Args:
            graph (nx.DiGraph): NetworkX directed graph with nodes as detections and
                edges as links.
            segmentation (np.ndarray | None): Optional segmentation array where labels
                match node IDs. Required for computing region properties (area, etc.).
            time_attr (str | None): Graph attribute name for time. Defaults to "time"
                if None.
            pos_attr (str | tuple[str, ...] | list[str] | None): Graph attribute
                name(s) for position. Can be:
                - Single string for one attribute containing position array
                - List/tuple of strings for multi-axis (one attribute per axis)
                Defaults to "pos" if None.
            tracklet_attr (str | None): Graph attribute name for tracklet/track IDs.
                Defaults to "track_id" if None.
            scale (list[float] | None): Scaling factors for each dimension (including
                time). If None, all dimensions scaled by 1.0.
            ndim (int | None): Number of dimensions (3 for 2D+time, 4 for 3D+time).
                If None, inferred from segmentation or scale.
            features (FeatureDict | None): Pre-built FeatureDict with feature
                definitions. If provided, time_attr/pos_attr/tracklet_attr are ignored.
                Assumes that all features in the dict already exist on the graph (will
                be activated but not recomputed). If None, core computed features (pos,
                area, track_id) are auto-detected by checking if they exist on the graph.
        """
        super().__init__(
            graph,
            segmentation=segmentation,
            time_attr=time_attr,
            pos_attr=pos_attr,
            tracklet_attr=tracklet_attr,
            scale=scale,
            ndim=ndim,
            features=features,
        )

        self.track_annotator = self._get_track_annotator()

    def _initialize_track_ids(self) -> None:
        """Initialize track IDs for all nodes.

        Deprecated:
            This method is deprecated and will be removed in funtracks v2.0.
            Track IDs are now auto-computed during SolutionTracks initialization.
        """
        warnings.warn(
            "`_initialize_track_ids` is deprecated and will be removed in funtracks v2.0."
            " Track IDs are now auto-computed during SolutionTracks initialization.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.enable_features([self.features.tracklet_key])  # type: ignore

    def _get_track_annotator(self) -> TrackAnnotator:
        """Get the TrackAnnotator instance from the annotator registry.

        Returns:
            TrackAnnotator: The track annotator instance

        Raises:
            RuntimeError: If no TrackAnnotator is registered
        """
        from funtracks.annotators import TrackAnnotator

        for annotator in self.annotators:
            if isinstance(annotator, TrackAnnotator):
                return annotator
        raise RuntimeError(
            "No TrackAnnotator registered for this SolutionTracks instance"
        )

    @classmethod
    def from_tracks(cls, tracks: Tracks):
        force_recompute = False
        if (tracklet_key := tracks.features.tracklet_key) is not None:
            # Check if all nodes have track_id before trusting existing track IDs
            # Short circuit on first missing track_id
            for node in tracks.graph.nodes():
                if tracks.get_node_attr(node, tracklet_key) is None:
                    force_recompute = True
                    break
        soln_tracks = cls(
            tracks.graph,
            segmentation=tracks.segmentation,
            scale=tracks.scale,
            ndim=tracks.ndim,
            features=tracks.features,
        )
        if force_recompute:
            soln_tracks.enable_features([soln_tracks.features.tracklet_key])  # type: ignore
        return soln_tracks

    @property
    def max_track_id(self) -> int:
        return self.track_annotator.max_tracklet_id

    @property
    def track_id_to_node(self) -> dict[int, list[int]]:
        return self.track_annotator.tracklet_id_to_nodes

    @property
    def node_id_to_track_id(self) -> dict[Node, int]:
        warnings.warn(
            "node_id_to_track_id property will be removed in funtracks v2. "
            "Use `get_track_id` instead for better performance.",
            DeprecationWarning,
            stacklevel=2,
        )
        return nx.get_node_attributes(self.graph, self.features.tracklet_key)

    def get_next_track_id(self) -> int:
        """Return the next available track_id and update max_tracklet_id in TrackAnnotator

        # TODO: I don't think we need to update the max here, it will get updated if we
        actually add a node automatically.
        """
        annotator = self.track_annotator
        annotator.max_tracklet_id = annotator.max_tracklet_id + 1
        return annotator.max_tracklet_id

    def get_track_id(self, node) -> int:
        if self.features.tracklet_key is None:
            raise ValueError("Tracklet key not initialized in features")
        track_id = self.get_node_attr(node, self.features.tracklet_key, required=True)
        return track_id

    def export_tracks(
        self, outfile: Path | str, node_ids: set[int] | None = None
    ) -> None:
        """Export the tracks from this run to a csv with the following columns:
        t,[z],y,x,id,parent_id,track_id
        Cells without a parent_id will have an empty string for the parent_id.
        Whether or not to include z is inferred from self.ndim

        Args:
            outfile (Path): path to output csv file
            node_ids (set[int], optional): nodes to be included. If provided, only these
            nodes and their ancestors will be included in the output.

        .. deprecated:: 1.0
            `SolutionTracks.export_tracks()` is deprecated and will be removed in v2.0.
            Use :func:`funtracks.import_export.export_to_csv` instead.
        """
        warnings.warn(
            "SolutionTracks.export_tracks() is deprecated and will be removed in v2.0. "
            "Use funtracks.import_export.export_to_csv() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        # Import here to avoid circular imports
        from funtracks.import_export.csv._export import export_to_csv

        export_to_csv(self, outfile, node_ids)

    def get_track_neighbors(
        self, track_id: int, time: int
    ) -> tuple[Node | None, Node | None]:
        """Get the last node with the given track id before time, and the first node
        with the track id after time, if any. Does not assume that a node with
        the given track_id and time is already in tracks, but it can be.

        Args:
            track_id (int): The track id to search for
            time (int): The time point to find the immediate predecessor and successor
                for

        Returns:
            tuple[Node | None, Node | None]: The last node before time with the given
            track id, and the first node after time with the given track id,
            or Nones if there are no such nodes.
        """
        annotator = self.track_annotator
        if (
            track_id not in annotator.tracklet_id_to_nodes
            or len(annotator.tracklet_id_to_nodes[track_id]) == 0
        ):
            return None, None
        candidates = annotator.tracklet_id_to_nodes[track_id]
        candidates.sort(key=lambda n: self.get_time(n))

        pred = None
        succ = None
        for cand in candidates:
            if self.get_time(cand) < time:
                pred = cand
            elif self.get_time(cand) > time:
                succ = cand
                break
        return pred, succ

    def has_track_id_at_time(self, track_id: int, time: int) -> bool:
        """Function to check if a node with given track id exists at given time point.

        Args:
            track_id (int): The track id to search for.
            time (int): The time point to check.

        Returns:
            True if a node with given track id exists at given time point.
        """

        nodes = self.track_id_to_node.get(track_id)
        if not nodes:
            return False

        return time in self.get_times(nodes)



import pytest

from funtracks.annotators import EdgeAnnotator, RegionpropsAnnotator, TrackAnnotator
from funtracks.data_model import SolutionTracks, Tracks

track_attrs = {"time_attr": "t", "tracklet_attr": "track_id"}


def test_annotator_registry_init_with_segmentation(graph_clean, segmentation_2d):
    """Test AnnotatorRegistry initializes regionprops and edge annotators with
    segmentation."""
    tracks = Tracks(graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs)

    annotator_types = [type(ann) for ann in tracks.annotators]
    assert RegionpropsAnnotator in annotator_types
    assert EdgeAnnotator in annotator_types
    assert TrackAnnotator not in annotator_types  # Not a SolutionTracks


def test_annotator_registry_init_without_segmentation(graph_2d_with_position):
    """Test AnnotatorRegistry doesn't create annotators without segmentation."""
    tracks = Tracks(graph_2d_with_position, segmentation=None, ndim=3, **track_attrs)

    annotator_types = [type(ann) for ann in tracks.annotators]
    assert RegionpropsAnnotator not in annotator_types
    assert EdgeAnnotator not in annotator_types
    assert TrackAnnotator not in annotator_types


def test_annotator_registry_init_solution_tracks(graph_clean, segmentation_2d):
    """Test AnnotatorRegistry creates all annotators for SolutionTracks with
    segmentation."""
    tracks = SolutionTracks(
        graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs
    )

    annotator_types = [type(ann) for ann in tracks.annotators]
    assert RegionpropsAnnotator in annotator_types
    assert EdgeAnnotator in annotator_types
    assert TrackAnnotator in annotator_types


def test_enable_disable_features(graph_clean, segmentation_2d):
    tracks = Tracks(graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs)

    nodes = list(tracks.graph.nodes())
    edges = list(tracks.graph.edges())

    # Core features (time, pos, area) should be in tracks.features and computed
    assert "pos" in tracks.features
    assert "t" in tracks.features
    assert "area" in tracks.features  # Core feature for backward compatibility
    assert tracks.graph.nodes[nodes[0]].get("pos") is not None
    assert tracks.graph.nodes[nodes[0]].get("area") is not None

    # Other features should NOT be in tracks.features initially
    assert "iou" not in tracks.features
    assert "circularity" not in tracks.features

    # Enable multiple features at once
    tracks.enable_features(["iou", "circularity"])

    # Features should now be in FeatureDict
    assert "iou" in tracks.features
    assert "circularity" in tracks.features

    # Verify values are actually computed on the graph
    assert tracks.graph.nodes[nodes[0]].get("circularity") is not None
    if edges:
        assert tracks.graph.edges[edges[0]].get("iou") is not None

    # Disable one feature
    tracks.disable_features(["area"])

    # area should be removed from FeatureDict
    assert "area" not in tracks.features
    assert "pos" in tracks.features
    assert "iou" in tracks.features
    assert "circularity" in tracks.features

    # Values still exist on the graph (disabling doesn't erase computed values)
    assert tracks.graph.nodes[nodes[0]].get("area") is not None

    # Disable the remaining enabled features
    tracks.disable_features(["pos", "iou", "circularity"])
    assert "pos" not in tracks.features
    assert "iou" not in tracks.features
    assert "circularity" not in tracks.features


def test_get_available_features(graph_clean, segmentation_2d):
    """Test get_available_features returns all features from all annotators."""
    tracks = SolutionTracks(
        graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs
    )

    available = tracks.get_available_features()

    # Should have features from all three annotators
    assert "pos" in available  # regionprops
    assert "area" in available  # regionprops
    assert "iou" in available  # edges
    assert "track_id" in available  # tracks


def test_enable_nonexistent_feature(graph_clean, segmentation_2d):
    """Test enabling a nonexistent feature raises KeyError."""
    tracks = Tracks(graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs)

    with pytest.raises(KeyError, match="Features not available"):
        tracks.enable_features(["nonexistent"])


def test_disable_nonexistent_feature(graph_clean, segmentation_2d):
    """Test disabling a nonexistent feature raises KeyError."""
    tracks = Tracks(graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs)

    with pytest.raises(KeyError, match="Features not available"):
        tracks.disable_features(["nonexistent"])


def test_compute_strict_validation(graph_clean, segmentation_2d):
    """Test that compute() strictly validates feature keys."""
    tracks = Tracks(graph_clean, segmentation=segmentation_2d, ndim=3, **track_attrs)

    # Get the RegionpropsAnnotator from the annotators
    rp_ann = next(
        ann for ann in tracks.annotators if isinstance(ann, RegionpropsAnnotator)
    )

    # Enable area first
    tracks.enable_features(["area"])

    # Valid feature key should work
    rp_ann.compute(["area"])

    # Invalid feature key should not raise KeyError
    rp_ann.compute(["nonexistent_feature"])

    # Disabled feature should not raise KeyError
    tracks.disable_features(["area"])
    rp_ann.compute(["area"])

    # None should still work (compute all enabled features)
    rp_ann.compute()



from .user_add_edge import UserAddEdge
from .user_add_node import UserAddNode
from .user_delete_edge import UserDeleteEdge
from .user_delete_node import UserDeleteNode
from .user_update_segmentation import UserUpdateSegmentation



import copy

import networkx as nx
import pytest
from numpy.testing import assert_array_almost_equal

from funtracks.actions import (
    ActionGroup,
    AddEdge,
    DeleteEdge,
)

iou_key = "iou"


@pytest.mark.parametrize("ndim", [3, 4])
@pytest.mark.parametrize("with_seg", [True, False])
def test_add_delete_edges(get_tracks, ndim, with_seg):
    tracks = get_tracks(ndim=ndim, with_seg=with_seg, is_solution=True)
    reference_graph = copy.deepcopy(tracks.graph)
    reference_seg = copy.deepcopy(tracks.segmentation)

    # Create an empty tracks with just nodes (no edges)
    node_graph = nx.create_empty_copy(tracks.graph, with_data=True)
    tracks.graph = node_graph

    edges = [(1, 2), (1, 3), (3, 4), (4, 5)]

    action = ActionGroup(tracks=tracks, actions=[AddEdge(tracks, edge) for edge in edges])
    # TODO: What if adding an edge that already exists?
    # TODO: test all the edge cases, invalid operations, etc. for all actions
    assert set(tracks.graph.nodes()) == set(reference_graph.nodes())
    if with_seg:
        for edge in tracks.graph.edges():
            assert tracks.graph.edges[edge][iou_key] == pytest.approx(
                reference_graph.edges[edge][iou_key], abs=0.01
            )
        assert_array_almost_equal(tracks.segmentation, reference_seg)

    inverse = action.inverse()
    assert set(tracks.graph.edges()) == set()
    if tracks.segmentation is not None:
        assert_array_almost_equal(tracks.segmentation, reference_seg)

    inverse.inverse()
    assert set(tracks.graph.nodes()) == set(reference_graph.nodes())
    assert set(tracks.graph.edges()) == set(reference_graph.edges())
    if with_seg:
        for edge in tracks.graph.edges():
            assert tracks.graph.edges[edge][iou_key] == pytest.approx(
                reference_graph.edges[edge][iou_key], abs=0.01
            )
        assert_array_almost_equal(tracks.segmentation, reference_seg)


def test_add_edge_missing_endpoint(get_tracks):
    tracks = get_tracks(ndim=3, with_seg=True, is_solution=True)
    with pytest.raises(ValueError, match="Cannot add edge .*: endpoint .* not in graph"):
        AddEdge(tracks, (10, 11))


def test_delete_missing_edge(get_tracks):
    tracks = get_tracks(ndim=3, with_seg=True, is_solution=True)
    with pytest.raises(
        ValueError, match="Edge .* not in the graph, and cannot be removed"
    ):
        DeleteEdge(tracks, (10, 11))


@pytest.mark.parametrize("ndim", [3, 4])
@pytest.mark.parametrize("with_seg", [True, False])
def test_custom_edge_attributes_preserved(get_tracks, ndim, with_seg):
    """Test custom edge attributes preserved through add/delete/re-add cycles."""
    from funtracks.features import Feature

    tracks = get_tracks(ndim=ndim, with_seg=with_seg, is_solution=True)

    # Register custom edge features so they get saved by DeleteEdge
    custom_features = {
        "edge_type": Feature(
            feature_type="edge",
            value_type="str",
            num_values=1,
            display_name="Edge Type",
            required=False,
            default_value=None,
        ),
        "confidence": Feature(
            feature_type="edge",
            value_type="float",
            num_values=1,
            display_name="Confidence",
            required=False,
            default_value=None,
        ),
        "weight": Feature(
            feature_type="edge",
            value_type="float",
            num_values=1,
            display_name="Weight",
            required=False,
            default_value=None,
        ),
    }
    for key, feature in custom_features.items():
        tracks.features[key] = feature

    # Define custom edge attributes
    custom_attrs = {
        "edge_type": "division",
        "confidence": 0.92,
        "weight": 1.5,
    }

    # Add an edge with custom attributes
    edge = (1, 2)
    action = AddEdge(tracks, edge, attributes=custom_attrs)

    # Verify all attributes are present after adding
    assert tracks.graph.has_edge(*edge)
    for key, value in custom_attrs.items():
        assert tracks.graph.edges[edge][key] == value, (
            f"Attribute {key} not preserved after add"
        )

    # Delete the edge
    delete_action = action.inverse()
    assert not tracks.graph.has_edge(*edge)

    # Re-add the edge by inverting the delete
    delete_action.inverse()
    assert tracks.graph.has_edge(*edge)

    # Verify all custom attributes are still present after re-adding
    for key, value in custom_attrs.items():
        assert tracks.graph.edges[edge][key] == value, (
            f"Attribute {key} not preserved after delete/re-add cycle"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ._base import BasicAction

if TYPE_CHECKING:
    from typing import Any

    from funtracks.data_model import Tracks
    from funtracks.data_model.tracks import Edge


class AddEdge(BasicAction):
    """Action for adding a new edge. Endpoints must exist already."""

    def __init__(
        self, tracks: Tracks, edge: Edge, attributes: dict[str, Any] | None = None
    ):
        """Create an action to add a new edge, with optional attributes.

        Args:
            tracks (Tracks): The Tracks to add the edge to
            edge (Edge): The edge to add (source, target)
            attributes (dict[str, Any] | None, optional): Edge attributes to set.
               If any computed features are provided, they will be overridden by the
               Annotator. Defaults to None.
        """
        super().__init__(tracks)
        self.edge = edge
        self.attributes = attributes if attributes is not None else {}
        self._apply()

    def inverse(self) -> BasicAction:
        """Delete edges"""
        return DeleteEdge(self.tracks, self.edge)

    def _apply(self) -> None:
        """
        Steps:
        - check for valid endpoints
        - add edge to the graph with provided attributes
        - Trigger feature (re)computation

        Raises:
            ValueError if an endpoint of the edge does not exist
        """
        # Check that both endpoints exist before computing edge attributes
        for node in self.edge:
            if not self.tracks.graph.has_node(node):
                raise ValueError(
                    f"Cannot add edge {self.edge}: endpoint {node} not in graph yet"
                )

        self.tracks.graph.add_edge(self.edge[0], self.edge[1], **self.attributes)

        # Notify annotators to recompute features (will overwrite computed ones)
        self.tracks.notify_annotators(self)


class DeleteEdge(BasicAction):
    """Action for deleting an edge. Edge must exist already."""

    def __init__(self, tracks: Tracks, edge: Edge):
        """Action for deleting an edge. Edge must exist already.

        Args:
            tracks (Tracks): The tracks to delete the edge from
            edge (Edge): The edge to delete
        Raises:
            ValueError: If the edge does not exist on the graph
        """
        super().__init__(tracks)
        self.edge = edge
        if not self.tracks.graph.has_edge(*self.edge):
            raise ValueError(f"Edge {self.edge} not in the graph, and cannot be removed")

        # Save all edge feature values from the features dict
        self.attributes = {}
        for key in self.tracks.features.edge_features:
            val = tracks.get_edge_attr(edge, key)
            if val is not None:
                self.attributes[key] = val

        self._apply()

    def inverse(self) -> BasicAction:
        """Restore edge and their attributes"""
        return AddEdge(self.tracks, self.edge, attributes=self.attributes)

    def _apply(self) -> None:
        self.tracks.graph.remove_edge(*self.edge)



from ._tracks_builder import TracksBuilder
from .csv._export import export_to_csv
from .csv._import import CSVTracksBuilder, tracks_from_df
from .geff._export import export_to_geff
from .geff._import import GeffTracksBuilder, import_from_geff
from .internal_format import load_tracks, save_tracks
from .magic_imread import magic_imread

__all__ = [
    "TracksBuilder",
    "CSVTracksBuilder",
    "GeffTracksBuilder",
    "import_from_geff",
    "tracks_from_df",
    "export_to_csv",
    "export_to_geff",
    "save_tracks",
    "load_tracks",
    "magic_imread",
]



from __future__ import annotations

from ._feature import Feature


def IoU() -> Feature:
    """A feature representing Intersection over Union for edges.

    Returns:
        Feature: A feature dict representing IoU
    """
    return {
        "feature_type": "edge",
        "value_type": "float",
        "num_values": 1,
        "display_name": "IoU",
        "required": True,
        "default_value": None,
    }



import dask.array as da
import numpy as np
import pytest
import tifffile
from geff.testing.data import create_mock_geff

from funtracks.import_export import import_from_geff
from funtracks.import_export.geff._import import import_graph_from_geff


@pytest.fixture
def valid_geff():
    store, memory_geff = create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "int64"},
        directed=True,
        num_nodes=5,
        num_edges=2,
        include_t=True,
        include_z=False,
        include_y=True,
        include_x=True,
        extra_node_props={
            "track_id": np.arange(5),
            "seg_id": np.array([10, 20, 30, 40, 50]),
            "lineage_id": np.arange(5),
            "area": np.array([20, 41, 42, 776, 21]),
            "circ": np.array([0.2, 0.1, 0.5, 0.3, 0.45]),
            "random_feature": np.array(["a", "b", "c", "d", "e"]),
            "random_feature2": np.array(["a", "b", "c", "d", "e"]),
        },
    )
    return store, memory_geff


@pytest.fixture
def invalid_geff():
    invalid_store, invalid_memory_geff = create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "int64"},
        directed=True,
        num_nodes=5,
        num_edges=2,
        include_t=True,
        include_z=False,
        include_y=True,
        include_x=True,
        extra_node_props={
            "track_id": np.arange(5),
            "seg_id": np.array([10.453, 20.23, 30.56, 40.78, 50.92]),
            "lineage_id": np.arange(5),
            "area": np.array([20, 41, 42, 776, 21]),
        },
    )
    return invalid_store, invalid_memory_geff


@pytest.fixture
def valid_segmentation():
    shape = (6, 600, 200)
    seg = np.zeros(shape, dtype=int)

    times = [1, 2, 3, 4, 5]
    x = [1.0, 0.775, 0.55, 0.325, 0.1]
    y = [100, 200, 300, 400, 500]
    scale = [1, 1, 100]
    seg_ids = np.array([10, 20, 30, 40, 50])

    for t, y_val, x_f, seg_id in zip(times, y, x, seg_ids, strict=False):
        x = int(x_f * scale[2])
        seg[t, y_val, x] = seg_id
    return seg


def test_import_graph_from_geff_renames_keys_to_standard(valid_geff):
    """Test that import_graph_from_geff renames custom GEFF keys to standard keys.

    This is a key architectural requirement: import_graph_from_geff should return
    an InMemoryGeff where all node_props keys have been renamed from custom GEFF
    property names to standard funtracks keys, using the provided node_name_map.
    """
    store, original_geff = valid_geff

    # Define node_name_map: standard_key -> custom_geff_key
    node_name_map = {
        "time": "t",  # standard key "time" maps to GEFF key "t"
        "y": "y",  # standard key "y" maps to GEFF key "y"
        "x": "x",  # standard key "x" maps to GEFF key "x"
        "track_id": "track_id",  # standard key "track_id" maps to GEFF key "track_id"
        "seg_id": "seg_id",  # standard key "seg_id" maps to GEFF key "seg_id"
        "circularity": "circ",  # standard key "circularity" maps to GEFF key "circ"
    }

    # Call import_graph_from_geff
    in_memory_geff, position_attr, ndims = import_graph_from_geff(store, node_name_map)

    # Assert the InMemoryGeff has standard keys, NOT custom GEFF keys
    node_props = in_memory_geff["node_props"]

    # Standard keys should be present
    assert "time" in node_props, "Standard key 'time' should be present"
    assert "y" in node_props, "Standard key 'y' should be present"
    assert "x" in node_props, "Standard key 'x' should be present"
    assert "track_id" in node_props, "Standard key 'track_id' should be present"
    assert "seg_id" in node_props, "Standard key 'seg_id' should be present"
    assert "circularity" in node_props, "Standard key 'circularity' should be present"

    # Custom GEFF keys should NOT be present
    assert "t" not in node_props, "Custom GEFF key 't' should have been renamed to 'time'"
    assert "circ" not in node_props, (
        "Custom GEFF key 'circ' should have been renamed to 'circularity'"
    )

    # Verify data integrity - values should be preserved
    assert len(node_props["time"]["values"]) == 5, "Should have 5 time values"
    assert len(node_props["track_id"]["values"]) == 5, "Should have 5 track_id values"
    np.testing.assert_array_equal(
        node_props["track_id"]["values"][:],
        np.arange(5),
        err_msg="track_id values should be preserved after renaming",
    )
    np.testing.assert_array_almost_equal(
        node_props["circularity"]["values"][:],
        np.array([0.2, 0.1, 0.5, 0.3, 0.45]),
        err_msg="circularity values should be preserved after renaming from 'circ'",
    )

    # Verify return values
    assert position_attr == ["y", "x"], "Should return standard position keys"
    assert ndims == 3, "Should be 3D (time + 2 spatial dims)"


def test_import_graph_from_geff_loads_custom_features(valid_geff):
    """Test that custom features can be loaded by including them in node_name_map.

    Custom features (not in the standard set) should be loaded when included
    in the node_name_map. The key should remain as the standard key (which equals
    the GEFF key in this case).
    """
    store, original_geff = valid_geff

    # Include custom features in node_name_map
    node_name_map = {
        "time": "t",
        "y": "y",
        "x": "x",
        "random_feature": "random_feature",  # Custom feature: maps to itself
        "random_feature2": "random_feature2",  # Another custom feature
    }

    # Call import_graph_from_geff
    in_memory_geff, position_attr, ndims = import_graph_from_geff(store, node_name_map)

    node_props = in_memory_geff["node_props"]

    # Custom features should be present with standard keys
    assert "random_feature" in node_props, (
        "Custom feature 'random_feature' should be loaded"
    )
    assert "random_feature2" in node_props, (
        "Custom feature 'random_feature2' should be loaded"
    )

    # Verify data integrity
    np.testing.assert_array_equal(
        node_props["random_feature"]["values"][:],
        np.array(["a", "b", "c", "d", "e"]),
        err_msg="random_feature values should be preserved",
    )
    np.testing.assert_array_equal(
        node_props["random_feature2"]["values"][:],
        np.array(["a", "b", "c", "d", "e"]),
        err_msg="random_feature2 values should be preserved",
    )


def test_import_graph_from_geff_custom_feature_with_different_name(valid_geff):
    """Test that custom features can be renamed using node_name_map.

    A custom feature with a GEFF name can be renamed to a different standard key
    using the node_name_map.
    """
    store, original_geff = valid_geff

    # Rename "circ" to "my_custom_circularity"
    node_name_map = {
        "time": "t",
        "y": "y",
        "x": "x",
        "my_custom_circularity": "circ",  # Rename circ to custom name
    }

    # Call import_graph_from_geff
    in_memory_geff, position_attr, ndims = import_graph_from_geff(store, node_name_map)

    node_props = in_memory_geff["node_props"]

    # The custom key should be present
    assert "my_custom_circularity" in node_props, (
        "Custom renamed feature should be present"
    )

    # The original GEFF key should NOT be present
    assert "circ" not in node_props, "Original GEFF key should be renamed"

    # Verify data integrity
    np.testing.assert_array_almost_equal(
        node_props["my_custom_circularity"]["values"][:],
        np.array([0.2, 0.1, 0.5, 0.3, 0.45]),
        err_msg="Values should be preserved after custom renaming",
    )


def test_import_graph_from_geff_edge_name_map_none(valid_geff):
    """Test that edge_name_map=None loads all edge properties.

    When edge_name_map is None, all edge properties should be loaded with their
    original GEFF names (no renaming).
    """
    store, original_geff = valid_geff

    # Define node name map
    node_name_map = {
        "time": "t",
        "y": "y",
        "x": "x",
    }

    # Call import_graph_from_geff without edge_name_map (defaults to None)
    in_memory_geff, position_attr, ndims = import_graph_from_geff(
        store, node_name_map, edge_name_map=None
    )

    # Should load successfully
    assert "node_props" in in_memory_geff
    assert "edge_props" in in_memory_geff
    # The fixture has no edge properties, so edge_props should be empty
    assert in_memory_geff["edge_props"] == {}


def test_none_in_name_map(valid_geff):
    """Test that None values in required t/y/x attributes are caught"""

    store, _ = valid_geff
    # None value for required field should raise error
    name_map = {"time": None, "y": "y", "x": "x"}
    with pytest.raises(ValueError, match="None values"):
        import_from_geff(store, name_map)


def test_duplicate_values_in_name_map(valid_geff):
    """Test that duplicate values in name_map are allowed."""
    store, _ = valid_geff

    # Duplicate values should be allowed - each standard key gets a copy of the data
    name_map = {"time": "t", "y": "y", "x": "x", "seg_id": "t"}

    # Should not raise - seg_id maps to same source as time
    tracks = import_from_geff(store, name_map)

    # Both time and seg_id should be present with same values
    for node_id in tracks.graph.nodes():
        assert tracks.get_node_attr(node_id, "seg_id") == tracks.get_node_attr(
            node_id, "time"
        )


def test_segmentation_axes_mismatch(valid_geff, tmp_path):
    """Test checking if number of dimensions match and if coordinates are within
    bounds."""

    store, _ = valid_geff
    name_map = {"time": "t", "y": "y", "x": "x", "seg_id": "seg_id"}

    # Provide a segmentation with wrong shape
    wrong_seg = np.zeros((2, 20, 200), dtype=np.uint16)
    seg_path = tmp_path / "wrong_seg.npy"
    tifffile.imwrite(seg_path, wrong_seg)
    with pytest.raises(ValueError, match="out of bounds"):
        import_from_geff(store, name_map, segmentation_path=seg_path)

    # Provide a segmentation with a different number of dimensions than the graph.
    wrong_seg = np.zeros((2, 20, 200, 200), dtype=np.uint16)
    seg_path = tmp_path / "wrong_seg2.npy"
    tifffile.imwrite(seg_path, wrong_seg)
    with pytest.raises(
        ValueError, match="Segmentation has 4 dimensions but graph has 3 dimensions"
    ):
        import_from_geff(store, name_map, segmentation_path=seg_path)


def test_tracks_with_segmentation(valid_geff, invalid_geff, valid_segmentation, tmp_path):
    """Test relabeling of the segmentation from seg_id to node_id."""

    store, _ = valid_geff
    name_map = {"time": "t", "y": "y", "x": "x", "seg_id": "seg_id"}
    valid_segmentation_path = tmp_path / "segmentation.tif"
    tifffile.imwrite(valid_segmentation_path, valid_segmentation)

    # Test that a tracks object is produced and that the seg_id has been relabeled.
    scale = [1, 1, (1 / 100)]
    node_features = {
        "area": True,  # In geff, but should be recomputed
        "random_feature": False,  # Static feature - load from geff
    }

    tracks = import_from_geff(
        store,
        name_map,
        segmentation_path=valid_segmentation_path,
        scale=scale,
        node_features=node_features,
    )
    assert hasattr(tracks, "segmentation")
    assert tracks.segmentation.shape == valid_segmentation.shape
    # Get last node by ID (don't rely on iteration order)
    last_node = max(tracks.graph.nodes())
    coords = [
        tracks.graph.nodes[last_node]["time"],
        tracks.graph.nodes[last_node]["y"],
        tracks.graph.nodes[last_node]["x"],
    ]
    coords = tuple(int(c * 1 / s) for c, s in zip(coords, scale, strict=True))
    assert (
        valid_segmentation[tuple(coords)] == 50
    )  # in original segmentation, the pixel value is equal to seg_id
    assert (
        tracks.segmentation[tuple(coords)] == last_node
    )  # test that the seg id has been relabeled

    # Check that only required/requested features are present, and that area is recomputed
    data = tracks.graph.nodes[last_node]
    assert "random_feature" in data
    assert "random_feature2" not in data
    assert "area" in data
    assert (
        data["area"] == 0.01
    )  # recomputed area values should be 1 pixel, so 0.01 after applying the scaling.

    # Check that area is not recomputed but taken directly from the graph
    node_features = {
        "area": False,  # Load from geff, don't recompute
        "random_feature": False,  # Static feature - load from geff
    }

    tracks = import_from_geff(
        store,
        name_map,
        segmentation_path=valid_segmentation_path,
        scale=scale,
        node_features=node_features,
    )
    # Get last node by ID (don't rely on iteration order)
    last_node = max(tracks.graph.nodes())
    data = tracks.graph.nodes[last_node]
    assert "area" in data
    assert data["area"] == 21

    # Test that import fails with ValueError when invalid seg_ids are provided.
    store, _ = invalid_geff
    with pytest.raises(ValueError):
        tracks = import_from_geff(
            store, name_map, segmentation_path=valid_segmentation_path, scale=scale
        )


@pytest.mark.parametrize("segmentation_format", ["single_tif", "tif_folder", "zarr"])
def test_segmentation_loading_formats(
    segmentation_format, valid_geff, valid_segmentation, tmp_path
):
    """Test loading segmentation from different formats using magic_imread."""
    store, _ = valid_geff
    name_map = {"time": "t", "y": "y", "x": "x", "seg_id": "seg_id"}
    scale = [1, 1, 1 / 100]
    seg = valid_segmentation

    if segmentation_format == "single_tif":
        path = tmp_path / "segmentation.tif"
        tifffile.imwrite(path, seg)

    elif segmentation_format == "tif_folder":
        path = tmp_path / "tif_series"
        path.mkdir()
        for i, frame in enumerate(seg):
            tifffile.imwrite(path / f"seg_{i:03}.tif", frame)

    elif segmentation_format == "zarr":
        path = tmp_path / "segmentation.zarr"
        da.from_array(seg, chunks=(1, *seg.shape[1:])).to_zarr(path)

    else:
        raise ValueError(f"Unknown format: {segmentation_format}")

    node_features = {
        "area": False,  # Load from geff, don't recompute
        "random_feature": False,  # Static feature - load from geff
    }

    tracks = import_from_geff(
        store,
        name_map,
        segmentation_path=path,
        scale=scale,
        node_features=node_features,
    )

    assert hasattr(tracks, "segmentation")
    assert np.array(tracks.segmentation).shape == seg.shape


def test_node_features_compute_vs_load(valid_geff, valid_segmentation, tmp_path):
    """Test that node_features controls whether features are computed or loaded.

    Features marked True in node_features are computed using annotators.
    Features marked False are loaded directly from the geff file.
    Features not in the geff can still be computed if marked True.
    """
    store, _ = valid_geff
    name_map = {
        "time": "t",
        "y": "y",
        "x": "x",
        "seg_id": "seg_id",
        "circularity": "circ",  # Map standard key to GEFF property name
    }
    scale = [1, 1, 1 / 100]
    valid_segmentation_path = tmp_path / "segmentation.tif"
    tifffile.imwrite(valid_segmentation_path, valid_segmentation)

    # Test 1: Mix of computed (True) and loaded (False) features
    node_features = {
        "area": True,  # In geff, but should be recomputed
        "random_feature": False,  # Static feature - load from geff
        "circularity": False,  # In geff, load without recomputing
        "ellipse_axis_radii": True,  # Not in geff, should be computed
    }

    tracks = import_from_geff(
        store,
        name_map,
        segmentation_path=valid_segmentation_path,
        scale=scale,
        node_features=node_features,
    )

    feature_keys = ["area", "random_feature", "ellipse_axis_radii", "circularity"]
    for key in feature_keys:
        assert key in tracks.features

    # Get last node by ID (don't rely on iteration order)
    max_node_id = max(tracks.graph.nodes())
    data = tracks.graph.nodes[max_node_id]

    # All requested features should be present
    for key in feature_keys:
        assert key in data

    # Verify computed values (1 pixel = 0.01 after scaling)
    # Original geff had area=21 for last node
    assert data["area"] == 0.01
    assert data["ellipse_axis_radii"] is not None
    assert data["circularity"] == 0.45  # the value should not be recomputed

    # Verify loaded value from geff
    assert data["random_feature"] == "e"


def test_node_features_unknown(valid_geff, valid_segmentation, tmp_path):
    """Test that providing an unknown feature raises a KeyError."""
    store, _ = valid_geff
    name_map = {"time": "t", "y": "y", "x": "x", "seg_id": "seg_id"}
    scale = [1, 1, 1 / 100]
    valid_segmentation_path = tmp_path / "segmentation.tif"
    tifffile.imwrite(valid_segmentation_path, valid_segmentation)

    # Test unknown feature that doesn't exist in annotators or GEFF
    node_features = {
        "unknown_computed_feature": True,  # Unknown feature, request computation
    }

    with pytest.raises(
        KeyError,
        match="Node features not available",
    ):
        import_from_geff(
            store,
            name_map,
            segmentation_path=valid_segmentation_path,
            scale=scale,
            node_features=node_features,
        )


def test_compute_features_without_segmentation(valid_geff):
    """Test that computing regionprops features without segmentation raises an error."""
    store, _ = valid_geff
    name_map = {"time": "t", "y": "y", "x": "x"}
    scale = [1, 1, 1 / 100]

    # Try to compute area feature without providing segmentation
    node_features = {
        "area": True,  # Request computation without segmentation
    }

    with pytest.raises(
        KeyError,
        match="Node features not available",
    ):
        import_from_geff(
            store,
            name_map,
            segmentation_path=None,  # No segmentation
            scale=scale,
            node_features=node_features,
        )



import networkx as nx

from funtracks.actions import AddNode
from funtracks.actions.action_history import ActionHistory
from funtracks.data_model import SolutionTracks

# https://github.com/zaboople/klonk/blob/master/TheGURQ.md


def test_action_history():
    history = ActionHistory()
    tracks = SolutionTracks(nx.DiGraph(), ndim=3, tracklet_attr="track_id")
    pos = [0, 1]
    action1 = AddNode(tracks, node=0, attributes={"time": 0, "pos": pos, "track_id": 1})

    # empty history has no undo or redo
    assert not history.undo()
    assert not history.redo()

    # add an action to the history
    history.add_new_action(action1)
    # undo the action
    assert history.undo()
    assert tracks.graph.number_of_nodes() == 0
    assert len(history.undo_stack) == 1
    assert len(history.redo_stack) == 1
    assert history._undo_pointer == -1

    # no more actions to undo
    assert not history.undo()

    # redo the action
    assert history.redo()
    assert tracks.graph.number_of_nodes() == 1
    assert len(history.undo_stack) == 1
    assert len(history.redo_stack) == 0
    assert history._undo_pointer == 0

    # no more actions to redo
    assert not history.redo()

    # undo and then add new action
    assert history.undo()
    action2 = AddNode(tracks, node=10, attributes={"time": 10, "pos": pos, "track_id": 2})
    history.add_new_action(action2)
    assert tracks.graph.number_of_nodes() == 1
    # there are 3 things on the stack: action1, action1's inverse, and action 2
    assert len(history.undo_stack) == 3
    assert len(history.redo_stack) == 0
    assert history._undo_pointer == 2

    # undo back to after action 1
    assert history.undo()
    assert history.undo()
    assert tracks.graph.number_of_nodes() == 1

    assert len(history.undo_stack) == 3
    assert len(history.redo_stack) == 2
    assert history._undo_pointer == 0



import copy
from collections.abc import Callable
from typing import TYPE_CHECKING

import networkx as nx
import numpy as np
import pytest
from skimage.draw import disk

if TYPE_CHECKING:
    from typing import Any

    from numpy.typing import NDArray

    from funtracks.data_model import SolutionTracks, Tracks

# Feature list constants for consistent test usage
FEATURES_WITH_SEG = ["pos", "area", "iou"]
FEATURES_NO_SEG = ["pos"]
SOLUTION_FEATURES_WITH_SEG = ["pos", "area", "iou", "track_id"]
SOLUTION_FEATURES_NO_SEG = ["pos", "track_id"]


@pytest.fixture
def segmentation_2d() -> "NDArray[np.int32]":
    frame_shape = (100, 100)
    total_shape = (5, *frame_shape)
    segmentation = np.zeros(total_shape, dtype="int32")
    # make frame with one cell in center with label 1
    rr, cc = disk(center=(50, 50), radius=20, shape=(100, 100))
    segmentation[0][rr, cc] = 1

    # make frame with two cells
    # first cell centered at (20, 80) with label 2
    # second cell centered at (60, 45) with label 3
    rr, cc = disk(center=(20, 80), radius=10, shape=frame_shape)
    segmentation[1][rr, cc] = 2
    rr, cc = disk(center=(60, 45), radius=15, shape=frame_shape)
    segmentation[1][rr, cc] = 3

    # continue track 3 with squares from 0 to 4 in x and y with label 3
    segmentation[2, 0:4, 0:4] = 4
    segmentation[4, 0:4, 0:4] = 5

    # unconnected node
    segmentation[4, 96:100, 96:100] = 6

    return segmentation


def _make_graph(
    *,
    ndim: int = 3,
    with_pos: bool = False,
    with_track_id: bool = False,
    with_area: bool = False,
    with_iou: bool = False,
) -> nx.DiGraph:
    """Generate a test graph with configurable features.

    Args:
        ndim: 3 for 2D spatial + time, 4 for 3D spatial + time
        with_pos: Include position attribute
        with_track_id: Include track_id attribute
        with_area: Include area attribute (requires with_pos=True)
        with_iou: Include iou edge attribute (requires with_area=True)

    Returns:
        A graph with the requested features
    """
    graph = nx.DiGraph()

    # Base node data (always has time)
    base_nodes = [
        (1, {"t": 0}),
        (2, {"t": 1}),
        (3, {"t": 1}),
        (4, {"t": 2}),
        (5, {"t": 4}),
        (6, {"t": 4}),
    ]

    # Position data
    if ndim == 3:  # 2D spatial
        positions = {
            1: [50, 50],
            2: [20, 80],
            3: [60, 45],
            4: [1.5, 1.5],
            5: [1.5, 1.5],
            6: [97.5, 97.5],
        }
        areas = {1: 1245, 2: 305, 3: 697, 4: 16, 5: 16, 6: 16}
        ious = {(1, 2): 0.0, (1, 3): 0.395, (3, 4): 0.0, (4, 5): 1.0}
    else:  # 3D spatial
        positions = {
            1: [50, 50, 50],
            2: [20, 50, 80],
            3: [60, 50, 45],
            4: [1.5, 1.5, 1.5],
            5: [1.5, 1.5, 1.5],
            6: [97.5, 97.5, 97.5],
        }
        areas = {1: 33401, 2: 4169, 3: 14147, 4: 64, 5: 64, 6: 64}
        ious = {(1, 2): 0.0, (1, 3): 0.302, (3, 4): 0.0, (4, 5): 1.0}

    # Track IDs
    track_ids = {1: 1, 2: 2, 3: 3, 4: 3, 5: 3, 6: 5}

    # Build nodes with requested features
    nodes = []
    for node_id, attrs in base_nodes:
        node_attrs: dict[str, Any] = dict(attrs)  # Start with time
        if with_pos:
            node_attrs["pos"] = positions[node_id]
        if with_track_id:
            node_attrs["track_id"] = track_ids[node_id]
        if with_area:
            node_attrs["area"] = areas[node_id]
        nodes.append((node_id, node_attrs))

    edges = [(1, 2), (1, 3), (3, 4), (4, 5)]

    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)

    # Add IOUs to edges if requested
    if with_iou:
        for edge, iou in ious.items():
            if edge in graph.edges:
                graph.edges[edge]["iou"] = iou

    return graph


@pytest.fixture
def graph_clean() -> nx.DiGraph:
    """Base graph with only time - no positions or computed features."""
    return _make_graph(ndim=3)


@pytest.fixture
def graph_2d_with_position() -> nx.DiGraph:
    """Graph with 2D positions - for Tracks without segmentation."""
    return _make_graph(ndim=3, with_pos=True)


@pytest.fixture
def graph_2d_with_track_id() -> nx.DiGraph:
    """Graph with 2D positions and track_id - for SolutionTracks without segmentation."""
    return _make_graph(ndim=3, with_pos=True, with_track_id=True)


@pytest.fixture
def graph_2d_with_computed_features() -> nx.DiGraph:
    """Graph with all computed features - for SolutionTracks with segmentation."""
    return _make_graph(
        ndim=3, with_pos=True, with_track_id=True, with_area=True, with_iou=True
    )


@pytest.fixture
def graph_3d_with_position() -> nx.DiGraph:
    """Graph with 3D positions - for Tracks without segmentation."""
    return _make_graph(ndim=4, with_pos=True)


@pytest.fixture
def graph_3d_with_track_id() -> nx.DiGraph:
    """Graph with 3D positions and track_id - for SolutionTracks without segmentation."""
    return _make_graph(ndim=4, with_pos=True, with_track_id=True)


@pytest.fixture
def graph_3d_with_computed_features() -> nx.DiGraph:
    """Graph with all computed features - for SolutionTracks with segmentation."""
    return _make_graph(
        ndim=4, with_pos=True, with_track_id=True, with_area=True, with_iou=True
    )


@pytest.fixture
def get_tracks(get_graph, get_segmentation) -> Callable[..., "Tracks | SolutionTracks"]:
    """Factory fixture to create Tracks or SolutionTracks instances.

    Returns a factory function that can be called with:
        ndim: 3 for 2D spatial + time, 4 for 3D spatial + time
        with_seg: Whether to include segmentation
        is_solution: Whether to return SolutionTracks instead of Tracks

    Example:
        tracks = get_tracks(ndim=3, with_seg=True, is_solution=True)

    Note:
        Uses a pre-built FeatureDict to avoid recomputing features that already
        exist in the test graph fixtures.
    """
    from funtracks.data_model import SolutionTracks, Tracks
    from funtracks.features import Area, FeatureDict, IoU, Position, Time, TrackletID

    def _make_tracks(
        ndim: int,
        with_seg: bool = True,
        is_solution: bool = False,
    ) -> Tracks | SolutionTracks:
        # Determine axis names based on ndim
        axis_names = ["z", "y", "x"] if ndim == 4 else ["y", "x"]

        # Determine which graph to use based on requirements
        if with_seg:
            # With segmentation: use fully computed features (pos + track_id + area + iou)
            graph = get_graph(ndim=ndim, with_features="computed")
            seg = get_segmentation(ndim=ndim)
        else:
            # Without segmentation
            if is_solution:
                # SolutionTracks needs track_id: use graph with pos + track_id
                graph = get_graph(ndim=ndim, with_features="track_id")
            else:
                # Regular Tracks: use graph with just pos
                graph = get_graph(ndim=ndim, with_features="position")
            seg = None

        # Build FeatureDict based on what exists in the graph
        features_dict = {
            "t": Time(),
            "pos": Position(axes=axis_names),
        }

        if with_seg:
            # Graph has pre-computed features (area, iou, track_id)
            features_dict["area"] = Area(ndim=ndim)
            features_dict["iou"] = IoU()
            features_dict["track_id"] = TrackletID()
        elif is_solution:
            # SolutionTracks without seg: has track_id but not area/iou
            features_dict["track_id"] = TrackletID()

        feature_dict = FeatureDict(
            features=features_dict,
            time_key="t",
            position_key="pos",
            tracklet_key="track_id" if (with_seg or is_solution) else None,
        )

        # Create the appropriate Tracks type with pre-built FeatureDict
        if is_solution:
            return SolutionTracks(
                graph,
                segmentation=seg,
                ndim=ndim,
                features=feature_dict,
            )
        else:
            return Tracks(
                graph,
                segmentation=seg,
                ndim=ndim,
                features=feature_dict,
            )

    return _make_tracks


@pytest.fixture
def graph_2d_list() -> nx.DiGraph:
    graph = nx.DiGraph()
    nodes = [
        (
            1,
            {
                "y": 100,
                "x": 50,
                "t": 0,
                "area": 1245,
                "track_id": 1,
            },
        ),
        (
            2,
            {
                "y": 20,
                "x": 100,
                "t": 1,
                "area": 500,
                "track_id": 2,
            },
        ),
    ]
    graph.add_nodes_from(nodes)
    return graph


def sphere(center, radius, shape):
    assert len(center) == len(shape)
    indices = np.moveaxis(np.indices(shape), 0, -1)  # last dim is the index
    distance = np.linalg.norm(np.subtract(indices, np.asarray(center)), axis=-1)
    mask = distance <= radius
    return mask


@pytest.fixture
def segmentation_3d() -> "NDArray[np.int32]":
    frame_shape = (100, 100, 100)
    total_shape = (5, *frame_shape)
    segmentation = np.zeros(total_shape, dtype="int32")
    # make frame with one cell in center with label 1
    mask = sphere(center=(50, 50, 50), radius=20, shape=frame_shape)
    segmentation[0][mask] = 1

    # make frame with two cells
    # first cell centered at (20, 50, 80) with label 2
    # second cell centered at (60, 50, 45) with label 3
    mask = sphere(center=(20, 50, 80), radius=10, shape=frame_shape)
    segmentation[1][mask] = 2
    mask = sphere(center=(60, 50, 45), radius=15, shape=frame_shape)
    segmentation[1][mask] = 3

    # continue track 3 with squares from 0 to 4 in x and y with label 3
    segmentation[2, 0:4, 0:4, 0:4] = 4
    segmentation[4, 0:4, 0:4, 0:4] = 5

    # unconnected node
    segmentation[4, 96:100, 96:100, 96:100] = 6
    return segmentation


@pytest.fixture
def get_graph(request) -> Callable[..., nx.DiGraph]:
    """Factory fixture to get graph by ndim and feature level.

    Args:
        ndim: 3 for 2D spatial + time, 4 for 3D spatial + time
        with_features: Feature level to include:
            - "clean": time only
            - "position": time + pos
            - "track_id": time + pos + track_id (for SolutionTracks without seg)
            - "computed": time + pos + track_id + area + iou (full features)

    Returns:
        A deep copy of the requested graph

    Example:
        graph = get_graph(ndim=3, with_features="track_id")
    """

    def _get_graph(ndim: int, with_features: str = "clean") -> nx.DiGraph:
        if with_features == "clean":
            graph = request.getfixturevalue("graph_clean")
        elif with_features == "position":
            if ndim == 3:
                graph = request.getfixturevalue("graph_2d_with_position")
            else:  # ndim == 4
                graph = request.getfixturevalue("graph_3d_with_position")
        elif with_features == "track_id":
            if ndim == 3:
                graph = request.getfixturevalue("graph_2d_with_track_id")
            else:  # ndim == 4
                graph = request.getfixturevalue("graph_3d_with_track_id")
        elif with_features == "computed":
            if ndim == 3:
                graph = request.getfixturevalue("graph_2d_with_computed_features")
            else:  # ndim == 4
                graph = request.getfixturevalue("graph_3d_with_computed_features")
        else:
            raise ValueError(
                f"with_features must be 'clean', 'position', 'track_id', or 'computed', "
                f"got {with_features}"
            )

        # Return a deep copy to avoid fixture pollution
        return copy.deepcopy(graph)

    return _get_graph


@pytest.fixture
def get_segmentation(request) -> Callable[..., "NDArray[np.int32]"]:
    """Factory fixture to get segmentation by ndim.

    Args:
        ndim: 3 for 2D spatial + time, 4 for 3D spatial + time

    Returns:
        The segmentation array (not copied since it's not typically modified)

    Example:
        seg = get_segmentation(ndim=3)
    """

    def _get_segmentation(ndim: int) -> "NDArray[np.int32]":
        if ndim == 3:
            return request.getfixturevalue("segmentation_2d")
        else:  # ndim == 4
            return request.getfixturevalue("segmentation_3d")

    return _get_segmentation



"""Zarr v2/v3 compatibility utilities.

This module provides utility functions for handling zarr version compatibility,
supporting both zarr-python v2 and v3, and both zarr spec v2 and v3.

Adapted from geff.core_io._utils.
"""

from __future__ import annotations

import os
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import zarr

if TYPE_CHECKING:
    from zarr.storage import StoreLike

# Type for zarr open modes
OpenMode = Literal["r", "r+", "a", "w", "w-"]

__all__ = [
    "detect_zarr_spec_version",
    "get_store_path",
    "is_zarr_v3",
    "open_zarr_store",
    "remove_tilde",
    "setup_zarr_array",
    "setup_zarr_group",
]


def is_zarr_v3() -> bool:
    """Check if the installed zarr-python version is v3.

    Returns:
        True if zarr-python v3 is installed, False otherwise.
    """
    return zarr.__version__.startswith("3")


def remove_tilde(store: StoreLike) -> StoreLike:
    """Remove tilde from a store path/str.

    zarr v3 does not recognize the tilde and may write the zarr
    in the wrong directory.

    Args:
        store: The store path to process

    Returns:
        The store with the tilde expanded
    """
    if isinstance(store, str | Path):
        store_str = str(store)
        if "~" in store_str:
            store = os.path.expanduser(store_str)
    return store


def detect_zarr_spec_version(store: StoreLike) -> int | None:
    """Detect the zarr specification version of an existing zarr store.

    Args:
        store: The zarr store path or object

    Returns:
        The zarr spec version (2 or 3) if detectable, None if unknown
    """
    try:
        if isinstance(store, str | Path):
            store_path = Path(store)
            # Check for zarr v3 indicator: zarr.json instead of .zarray/.zgroup
            if (store_path / "zarr.json").exists():
                return 3
            # Check for zarr v2 indicators
            elif (store_path / ".zgroup").exists() or (store_path / ".zarray").exists():
                return 2
        else:
            # For store objects, try to detect based on metadata
            group = zarr.open_group(store, mode="r")
            if group.metadata.zarr_format == 3:  # type: ignore[union-attr]
                return 3
            elif group.metadata.zarr_format == 2:  # type: ignore[union-attr]
                return 2
    except (OSError, KeyError, ValueError, AttributeError):
        # If we can't detect, return None
        pass

    return None


def setup_zarr_group(
    store: StoreLike, zarr_format: Literal[2, 3] = 2, mode: OpenMode = "a"
) -> zarr.Group:
    """Set up and return a zarr group for writing.

    Args:
        store: The zarr store path or object
        zarr_format: The zarr format version to use (default: 2)
        mode: The mode to open the group with (default: "a" for append)

    Returns:
        The opened zarr group
    """
    store = remove_tilde(store)

    # Check for trying to write zarr spec v3 with zarr python v2 and warn
    if zarr_format == 3 and not is_zarr_v3():
        warnings.warn(
            "Requesting zarr spec v3 with zarr-python v2. "
            "zarr-python v2 does not support spec v3. "
            "Ignoring zarr_format=3 and writing zarr spec v2 instead. "
            "Consider upgrading to zarr-python v3 to write zarr spec v3 files.",
            UserWarning,
            stacklevel=2,
        )

    # open/create zarr container
    if is_zarr_v3():
        return zarr.open_group(store, mode=mode, zarr_format=zarr_format)
    else:
        return zarr.open_group(store, mode=mode)


def setup_zarr_array(
    store: StoreLike,
    zarr_format: Literal[2, 3] = 2,
    mode: OpenMode = "w",
    **kwargs,
) -> zarr.Array:
    """Set up and return a zarr array for writing.

    Args:
        store: The zarr store path or object
        zarr_format: The zarr format version to use (default: 2)
        mode: The mode to open the array with (default: "w")
        **kwargs: Additional arguments passed to zarr.open_array
            (shape, dtype, chunks, etc.)

    Returns:
        The opened zarr array
    """
    store = remove_tilde(store)

    # Check for trying to write zarr spec v3 with zarr python v2 and warn
    if zarr_format == 3 and not is_zarr_v3():
        warnings.warn(
            "Requesting zarr spec v3 with zarr-python v2. "
            "zarr-python v2 does not support spec v3. "
            "Ignoring zarr_format=3 and writing zarr spec v2 instead. "
            "Consider upgrading to zarr-python v3 to write zarr spec v3 files.",
            UserWarning,
            stacklevel=2,
        )

    # open/create zarr array
    if is_zarr_v3():
        return zarr.open_array(store, mode=mode, zarr_format=zarr_format, **kwargs)
    else:
        return zarr.open_array(store, mode=mode, **kwargs)


def open_zarr_store(path: str | Path):
    """Open a zarr store with version-appropriate store class.

    In zarr-python v3, uses LocalStore. In v2, uses FSStore.

    Args:
        path: Path to the zarr store

    Returns:
        Appropriate store object for the installed zarr version

    Raises:
        FileNotFoundError: Path does not exist
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Path does not exist: {path}")

    # Check for zarr spec v3 files being opened with zarr python v2 and warn
    if not is_zarr_v3():
        spec_version = detect_zarr_spec_version(path)
        if spec_version == 3:
            warnings.warn(
                "Attempting to open a zarr spec v3 file with zarr-python v2. "
                "This may cause compatibility issues. Consider upgrading to "
                "zarr-python v3 or recreating the file with zarr spec v2.",
                UserWarning,
                stacklevel=2,
            )

    if is_zarr_v3():
        return zarr.storage.LocalStore(path)
    else:
        # FSStore only exists in zarr v2
        return zarr.storage.FSStore(str(path))  # type: ignore[attr-defined]


def get_store_path(store) -> Path:
    """Get the filesystem path from a zarr store.

    Handles differences between zarr v2 FSStore (.path) and v3 LocalStore (.root).

    Args:
        store: A zarr store object

    Returns:
        The filesystem path of the store

    Raises:
        ValueError: Cannot determine store path
    """
    if hasattr(store, "root"):
        return Path(store.root)
    elif hasattr(store, "path"):
        return Path(store.path)
    else:
        raise ValueError(f"Cannot determine store path from {type(store)}")



from __future__ import annotations

from typing import TYPE_CHECKING

from ._graph_annotator import GraphAnnotator

if TYPE_CHECKING:
    from funtracks.actions import BasicAction
    from funtracks.features import Feature


class AnnotatorRegistry(list[GraphAnnotator]):
    """A list of annotators with coordinated operations.

    Inherits from list[GraphAnnotator], so can be used directly as a list.
    Provides coordinated compute/update/enable/disable operations across all annotators.

    Example:
        annotators = AnnotatorRegistry([
            RegionpropsAnnotator(tracks, pos_key="centroid"),
            EdgeAnnotator(tracks),
            TrackAnnotator(tracks, tracklet_key="track_id"),
        ])

        # Can use as a list
        annotators.append(MyCustomAnnotator(tracks))

        # Coordinated operations
        annotators.activate_features(["area", "iou"])
        annotators.compute()
    """

    def __init__(self, annotators: list[GraphAnnotator]):
        """Initialize with a list of annotators.

        Args:
            annotators: List of instantiated annotator objects
        """
        super().__init__(annotators)

    @property
    def all_features(self) -> dict[str, tuple[Feature, bool]]:
        """Dynamically aggregate all_features from all annotators.

        Returns:
            Dictionary mapping feature keys to (Feature, is_enabled) tuples
        """
        aggregated = {}
        for annotator in self:
            aggregated.update(annotator.all_features)
        return aggregated

    @property
    def features(self) -> dict[str, Feature]:
        """Get all currently active features from all annotators.

        Returns:
            Dictionary mapping feature keys to Feature definitions (only active features)
        """
        aggregated = {}
        for annotator in self:
            aggregated.update(annotator.features)
        return aggregated

    def compute(self, feature_keys: list[str] | None = None) -> None:
        """Compute features across all annotators.

        Args:
            feature_keys: Optional list of specific feature keys to compute.
                If None, computes all currently active features.
        """
        for annotator in self:
            annotator.compute(feature_keys)

    def update(self, action: BasicAction) -> None:
        """Update features across all annotators based on the action.

        Args:
            action: The action that triggered this update
        """
        for annotator in self:
            annotator.update(action)

    def activate_features(self, keys: list[str]) -> None:
        """Activate features across all annotators.

        Args:
            keys: List of feature keys to activate

        Raises:
            KeyError: If any feature keys are not available
        """
        # Validate first - fail before making any changes
        available = self.all_features
        not_found = [k for k in keys if k not in available]
        if not_found:
            raise KeyError(f"Features not available: {not_found}")

        # All features exist - proceed with activating
        for annotator in self:
            annotator.activate_features(keys)

    def deactivate_features(self, keys: list[str]) -> None:
        """Deactivate features across all annotators.

        Args:
            keys: List of feature keys to deactivate

        Raises:
            KeyError: If any feature keys are not available
        """
        # Validate first - fail before making any changes
        available = self.all_features
        not_found = [k for k in keys if k not in available]
        if not_found:
            raise KeyError(f"Features not available: {not_found}")

        # All features exist - proceed with deactivation
        for annotator in self:
            annotator.deactivate_features(keys)

    def change_key(self, old_key: str, new_key: str) -> None:
        """Rename a feature key across all annotators.

        Finds the annotator that owns the feature and calls its change_key method.
        This allows the Tracks user to rename features without needing to find
        the specific annotator that manages it.

        Args:
            old_key: Existing key to rename.
            new_key: New key to replace it with.

        Raises:
            KeyError: If old_key does not exist in any annotator.
        """
        # Find which annotator owns this feature
        for annotator in self:
            if old_key in annotator.all_features:
                annotator.change_key(old_key, new_key)
                return

        # Key not found in any annotator
        raise KeyError(f"Cannot rename missing feature key: {old_key!r}")



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from funtracks.exceptions import InvalidActionError

from ..actions._base import ActionGroup
from ..actions.add_delete_edge import AddEdge
from ..actions.update_track_id import UpdateTrackID
from .user_delete_edge import UserDeleteEdge

if TYPE_CHECKING:
    from funtracks.data_model import SolutionTracks


class UserAddEdge(ActionGroup):
    """Assumes that the endpoints already exist and have track ids.

    Args:
        tracks (SolutionTracks): the tracks to add the edge to
        edge (tuple[int, int]): The edge to add
        force (bool, optional): Whether to force the action by removing any conflicting
            edges. Defaults to False.
    """

    def __init__(
        self,
        tracks: SolutionTracks,
        edge: tuple[int, int],
        force: bool = False,
    ):
        super().__init__(tracks, actions=[])
        self.tracks: SolutionTracks  # Narrow type from base class
        source, target = edge
        if not tracks.graph.has_node(source):
            raise InvalidActionError(
                f"Source node {source} not in solution yet - must be added before edge"
            )
        if not tracks.graph.has_node(target):
            raise InvalidActionError(
                f"Target node {target} not in solution yet - must be added before edge"
            )

        # Check if making a merge. If yes and force, remove the other edge and update
        # track ids.
        in_degree_target = self.tracks.graph.in_degree(target)
        if in_degree_target > 0:
            if not force:
                raise InvalidActionError(
                    f"Cannot make a merge edge in a tracking solution: node {target} "
                    "already has an in edge",
                    forceable=True,
                )
            else:
                merge_edge = list(self.tracks.graph.in_edges(target))[0]
                warnings.warn(
                    f"Removing edge {merge_edge} to add new edge without merging.",
                    stacklevel=2,
                )
                self.actions.append(UserDeleteEdge(self.tracks, merge_edge))

        # update track ids if needed
        out_degree_source = self.tracks.graph.out_degree(source)
        if out_degree_source == 0:  # joining two segments
            # assign the track id of the source node to the target and all out
            # edges until end of track
            new_track_id = self.tracks.get_track_id(source)
            self.actions.append(UpdateTrackID(self.tracks, edge[1], new_track_id))
        elif out_degree_source == 1:  # creating a division
            # assign a new track id to existing child
            successor = next(iter(self.tracks.graph.successors(source)))
            self.actions.append(
                UpdateTrackID(self.tracks, successor, self.tracks.get_next_track_id())
            )
        else:
            raise InvalidActionError(
                f"Expected degree of 0 or 1 before adding edge, got {out_degree_source}"
            )

        self.actions.append(AddEdge(tracks, edge))



from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import override

if TYPE_CHECKING:
    from funtracks.data_model import Tracks


class Action:
    """Base class for all actions that can be applied to Tracks.

    Two types of actions exist:
    - BasicAction: Atomic operations that directly modify tracks and trigger annotations
    - ActionGroup: Composite operations that contain and execute BasicActions
    """

    def __init__(self, tracks: Tracks):
        """A modular change that can be applied to the given Tracks. The tracks must
        be passed in at construction time so that metadata needed to invert the action
        can be extracted.
        The change should be applied in the init function.

        Args:
            tracks (Tracks): The tracks that this action will edit
        """
        self.tracks = tracks

    def inverse(self) -> Action:
        """Get the inverse of this action. Calling this function does undo the action,
        since the change is applied in the action constructor.

        Raises:
            NotImplementedError: if the inverse is not implemented in the subclass

        Returns:
            Action: An action that un-does this action, bringing the tracks
                back to the exact state it had before applying this action.
        """
        raise NotImplementedError("Inverse not implemented")


class BasicAction(Action):
    """Atomic action that directly modifies tracks and triggers annotation updates.

    BasicActions are the primitive operations that annotators listen to and respond to.
    Examples: AddNode, DeleteEdge, UpdateNodeSeg, etc.
    """


class ActionGroup(Action):
    """Composite action that contains and executes multiple Actions.

    ActionGroups are high-level operations that encapsulate application logic.
    They can contain BasicActions or other ActionGroups.
    They are not passed to annotators - only the BasicActions they contain are.
    Examples: UserAddNode, UserDeleteEdge, etc.
    """

    def __init__(
        self,
        tracks: Tracks,
        actions: list[Action],
    ):
        """A group of actions that is also an action.

        This is useful for creating composite actions from atomic BasicActions or other
        ActionGroups. Composite actions can contain application logic and can be un-done
        as a group.

        Args:
            tracks (Tracks): The tracks that this action will edit
            actions (list[Action]): A list of actions contained within the group, in the
                order in which they should be executed. Can be BasicActions or
                ActionGroups.
        """
        super().__init__(tracks)
        self.actions = actions

    @override
    def inverse(self) -> ActionGroup:
        actions = [action.inverse() for action in self.actions[::-1]]
        return ActionGroup(self.tracks, actions)



from collections import Counter

import numpy as np
import pytest

from funtracks.user_actions import UserUpdateSegmentation

iou_key = "iou"
area_key = "area"


# TODO: add area to the 4d testing graph
@pytest.mark.parametrize(
    "ndim",
    [3],
)
class TestUpdateNodeSeg:
    def test_user_update_seg_smaller(self, get_tracks, ndim):
        tracks = get_tracks(ndim=ndim, with_seg=True, is_solution=True)
        node_id = 3
        edge = (1, 3)

        orig_pixels = tracks.get_pixels(node_id)
        orig_position = tracks.get_position(node_id)
        orig_area = tracks.get_node_attr(node_id, area_key)
        orig_iou = tracks.get_edge_attr(edge, iou_key)

        # remove all but one pixel
        pixels_to_remove = tuple(orig_pixels[d][1:] for d in range(len(orig_pixels)))
        remaining_loc = tuple(orig_pixels[d][0] for d in range(len(orig_pixels)))
        new_position = [remaining_loc[1].item(), remaining_loc[2].item()]
        remaining_pixels = tuple(
            np.array([remaining_loc[d]]) for d in range(len(orig_pixels))
        )

        action = UserUpdateSegmentation(
            tracks,
            new_value=0,
            updated_pixels=[(pixels_to_remove, node_id)],
            current_track_id=1,
        )
        assert tracks.graph.has_node(node_id)
        assert self.pixel_equals(tracks.get_pixels(node_id), remaining_pixels)
        assert tracks.get_position(node_id) == new_position
        assert tracks.get_node_attr(node_id, "area") == 1
        assert tracks.get_edge_attr(edge, iou_key) == pytest.approx(0.0, abs=0.01)

        inverse = action.inverse()
        assert tracks.graph.has_node(node_id)
        assert self.pixel_equals(tracks.get_pixels(node_id), orig_pixels)
        assert tracks.get_position(node_id) == orig_position
        assert tracks.get_node_attr(node_id, "area") == orig_area
        assert tracks.get_edge_attr(edge, iou_key) == pytest.approx(orig_iou, abs=0.01)

        inverse.inverse()
        assert self.pixel_equals(tracks.get_pixels(node_id), remaining_pixels)
        assert tracks.get_position(node_id) == new_position
        assert tracks.get_node_attr(node_id, "area") == 1
        assert tracks.get_edge_attr(edge, iou_key) == pytest.approx(0.0, abs=0.01)

    def pixel_equals(self, pixels1, pixels2):
        return Counter(zip(*pixels1, strict=True)) == Counter(zip(*pixels2, strict=True))

    def test_user_update_seg_bigger(self, get_tracks, ndim):
        tracks = get_tracks(ndim=ndim, with_seg=True, is_solution=True)
        node_id = 3
        edge = (1, 3)

        orig_pixels = tracks.get_pixels(node_id)
        orig_position = tracks.get_position(node_id)
        orig_area = tracks.get_node_attr(node_id, "area")
        orig_iou = tracks.get_edge_attr(edge, iou_key)

        # add one pixel
        pixels_to_add = tuple(
            np.array([orig_pixels[d][0]]) for d in range(len(orig_pixels))
        )
        new_x_val = 10
        pixels_to_add = (*pixels_to_add[:-1], np.array([new_x_val]))
        all_pixels = tuple(
            np.concat([orig_pixels[d], pixels_to_add[d]]) for d in range(len(orig_pixels))
        )

        action = UserUpdateSegmentation(
            tracks, new_value=3, updated_pixels=[(pixels_to_add, 0)], current_track_id=1
        )
        assert tracks.graph.has_node(node_id)
        assert self.pixel_equals(all_pixels, tracks.get_pixels(node_id))
        assert tracks.get_node_attr(node_id, "area") == orig_area + 1
        assert tracks.get_edge_attr(edge, iou_key) != orig_iou

        inverse = action.inverse()
        assert tracks.graph.has_node(node_id)
        assert self.pixel_equals(orig_pixels, tracks.get_pixels(node_id))
        assert tracks.get_position(node_id) == orig_position
        assert tracks.get_node_attr(node_id, "area") == orig_area
        assert tracks.get_edge_attr(edge, iou_key) == pytest.approx(orig_iou, abs=0.01)

        inverse.inverse()
        assert tracks.graph.has_node(node_id)
        assert self.pixel_equals(all_pixels, tracks.get_pixels(node_id))
        assert tracks.get_node_attr(node_id, "area") == orig_area + 1
        assert tracks.get_edge_attr(edge, iou_key) != orig_iou

    def test_user_erase_seg(self, get_tracks, ndim):
        tracks = get_tracks(ndim=ndim, with_seg=True, is_solution=True)
        node_id = 3
        edge = (1, 3)

        orig_pixels = tracks.get_pixels(node_id)
        orig_position = tracks.get_position(node_id)
        orig_area = tracks.get_node_attr(node_id, "area")
        orig_iou = tracks.get_edge_attr(edge, iou_key)

        # remove all pixels
        pixels_to_remove = orig_pixels
        # set the pixels in the array first
        # (to reflect that the user directly changes the segmentation array)
        tracks.set_pixels(pixels_to_remove, 0)
        action = UserUpdateSegmentation(
            tracks,
            new_value=0,
            updated_pixels=[(pixels_to_remove, node_id)],
            current_track_id=1,
        )
        assert not tracks.graph.has_node(node_id)

        tracks.set_pixels(pixels_to_remove, node_id)
        inverse = action.inverse()
        assert tracks.graph.has_node(node_id)
        self.pixel_equals(tracks.get_pixels(node_id), orig_pixels)
        assert tracks.get_position(node_id) == orig_position
        assert tracks.get_node_attr(node_id, "area") == orig_area
        assert tracks.get_edge_attr(edge, iou_key) == pytest.approx(orig_iou, abs=0.01)

        tracks.set_pixels(pixels_to_remove, 0)
        inverse.inverse()
        assert not tracks.graph.has_node(node_id)

    def test_user_add_seg(self, get_tracks, ndim):
        tracks = get_tracks(ndim=ndim, with_seg=True, is_solution=True)
        # draw a new node just like node 6 but in time 3 (instead of 4)
        old_node_id = 6
        node_id = 7
        time = 3

        pixels_to_add = tracks.get_pixels(old_node_id)
        pixels_to_add = (
            np.ones(shape=(pixels_to_add[0].shape), dtype=np.uint32) * time,
            *pixels_to_add[1:],
        )
        position = tracks.get_position(old_node_id)
        area = tracks.get_node_attr(old_node_id, "area")

        assert not tracks.graph.has_node(node_id)

        assert np.sum(tracks.segmentation == node_id) == 0
        tracks.set_pixels(pixels_to_add, node_id)
        action = UserUpdateSegmentation(
            tracks,
            new_value=node_id,
            updated_pixels=[(pixels_to_add, 0)],
            current_track_id=10,
        )
        assert np.sum(tracks.segmentation == node_id) == len(pixels_to_add[0])
        assert tracks.graph.has_node(node_id)
        assert tracks.get_position(node_id) == position
        assert tracks.get_node_attr(node_id, "area") == area
        assert tracks.get_track_id(node_id) == 10

        inverse = action.inverse()
        assert not tracks.graph.has_node(node_id)

        inverse.inverse()
        assert tracks.graph.has_node(node_id)
        assert tracks.get_position(node_id) == position
        assert tracks.get_node_attr(node_id, "area") == area
        assert tracks.get_track_id(node_id) == 10


def test_missing_seg(get_tracks):
    tracks = get_tracks(ndim=3, with_seg=False, is_solution=True)
    with pytest.raises(ValueError, match="Cannot update non-existing segmentation"):
        UserUpdateSegmentation(tracks, 0, [], 1)



import networkx as nx
import numpy as np
import pytest
from networkx.utils import graphs_equal
from numpy.testing import assert_array_almost_equal

from funtracks.data_model import Tracks

track_attrs = {"time_attr": "t", "tracklet_attr": "track_id"}


def test_create_tracks(graph_3d_with_computed_features: nx.DiGraph, segmentation_3d):
    # create empty tracks
    tracks = Tracks(graph=nx.DiGraph(), ndim=3, **track_attrs)  # type: ignore[arg-type]
    assert tracks.features.position_key == "pos"
    assert isinstance(tracks.features["pos"], dict)
    with pytest.raises(KeyError):
        tracks.get_positions([1])

    # create tracks with graph only
    tracks = Tracks(
        graph=graph_3d_with_computed_features,
        ndim=4,
        **track_attrs,  # type: ignore[arg-type]
    )
    pos_key = tracks.features.position_key
    assert pos_key == "pos"
    assert isinstance(pos_key, str)
    assert isinstance(tracks.features[pos_key], dict)
    assert tracks.get_positions([1]).tolist() == [[50, 50, 50]]
    assert tracks.get_time(1) == 0
    with pytest.raises(KeyError):
        tracks.get_position(0)

    # create track with graph and seg
    tracks = Tracks(
        graph=graph_3d_with_computed_features,
        segmentation=segmentation_3d,
        **track_attrs,  # type: ignore[arg-type]
    )
    pos_key = tracks.features.position_key
    assert pos_key == "pos"
    assert isinstance(pos_key, str)
    assert isinstance(tracks.features[pos_key], dict)
    assert tracks.get_positions([1]).tolist() == [[50, 50, 50]]
    assert tracks.get_time(1) == 0
    assert tracks.get_positions([1], incl_time=True).tolist() == [[0, 50, 50, 50]]
    tracks.set_time(1, 1)
    assert tracks.get_positions([1], incl_time=True).tolist() == [[1, 50, 50, 50]]

    tracks_wrong_attr = Tracks(
        graph=graph_3d_with_computed_features,
        segmentation=segmentation_3d,
        time_attr="test",
    )
    with pytest.raises(KeyError):  # raises error at access if time is wrong
        tracks_wrong_attr.get_times([1])

    tracks_wrong_attr = Tracks(
        graph=graph_3d_with_computed_features, pos_attr="test", ndim=3
    )
    with pytest.raises(KeyError):  # raises error at access if pos is wrong
        tracks_wrong_attr.get_positions([1])

    # test multiple position attrs
    pos_attr = ("z", "y", "x")
    for node in graph_3d_with_computed_features.nodes():
        pos = graph_3d_with_computed_features.nodes[node]["pos"]
        z, y, x = pos
        del graph_3d_with_computed_features.nodes[node]["pos"]
        graph_3d_with_computed_features.nodes[node]["z"] = z
        graph_3d_with_computed_features.nodes[node]["y"] = y
        graph_3d_with_computed_features.nodes[node]["x"] = x

    tracks = Tracks(
        graph=graph_3d_with_computed_features,
        pos_attr=pos_attr,
        ndim=4,
        **track_attrs,  # type: ignore[arg-type]
    )
    assert tracks.get_positions([1]).tolist() == [[50, 50, 50]]
    tracks.set_position(1, [55, 56, 57])
    assert tracks.get_position(1) == [55, 56, 57]

    tracks.set_position(1, [1, 50, 50, 50], incl_time=True)
    assert tracks.get_time(1) == 1


def test_pixels_and_seg_id(graph_3d_with_computed_features, segmentation_3d):
    # create track with graph and seg
    tracks = Tracks(
        graph=graph_3d_with_computed_features, segmentation=segmentation_3d, **track_attrs
    )

    # changing a segmentation id changes it in the mapping
    pix = tracks.get_pixels(1)
    new_seg_id = 10
    tracks.set_pixels(pix, new_seg_id)


def test_save_load_delete(tmp_path, graph_2d_with_computed_features, segmentation_2d):
    tracks_dir = tmp_path / "tracks"
    tracks = Tracks(graph_2d_with_computed_features, segmentation_2d, **track_attrs)
    with pytest.warns(
        DeprecationWarning,
        match="`Tracks.save` is deprecated and will be removed in 2.0",
    ):
        tracks.save(tracks_dir)
    with pytest.warns(
        DeprecationWarning,
        match="`Tracks.load` is deprecated and will be removed in 2.0",
    ):
        loaded = Tracks.load(tracks_dir)
        assert graphs_equal(loaded.graph, tracks.graph)
        assert_array_almost_equal(loaded.segmentation, tracks.segmentation)
    with pytest.warns(
        DeprecationWarning,
        match="`Tracks.delete` is deprecated and will be removed in 2.0",
    ):
        Tracks.delete(tracks_dir)


def test_nodes_edges(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    assert set(tracks.nodes()) == {1, 2, 3, 4, 5, 6}
    assert set(map(tuple, tracks.edges())) == {(1, 2), (1, 3), (3, 4), (4, 5)}


def test_degrees(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    assert tracks.in_degree(np.array([1])) == 0
    assert tracks.in_degree(np.array([4])) == 1
    assert np.array_equal(
        tracks.in_degree(None), np.array([[1, 0], [2, 1], [3, 1], [4, 1], [5, 1], [6, 0]])
    )
    assert np.array_equal(tracks.out_degree(np.array([1, 4])), np.array([2, 1]))
    assert np.array_equal(
        tracks.out_degree(None),
        np.array([[1, 2], [2, 0], [3, 1], [4, 1], [5, 0], [6, 0]]),
    )


def test_predecessors_successors(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    assert tracks.predecessors(2) == [1]
    assert tracks.successors(1) == [2, 3]
    assert tracks.predecessors(1) == []
    assert tracks.successors(2) == []


def test_area_methods(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    assert tracks.get_area(1) == 1245
    assert tracks.get_areas([1, 2]) == [1245, 305]


def test_iou_methods(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    assert tracks.get_iou((1, 2)) == 0.0
    assert tracks.get_ious([(1, 2)]) == [0.0]
    assert tracks.get_ious([(1, 2), (1, 3)]) == [0.0, 0.395]


def test_get_set_node_attr(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)

    tracks._set_node_attr(1, "a", 42)
    # test deprecated functions
    with pytest.warns(
        DeprecationWarning,
        match="_get_node_attr deprecated in favor of public method get_node_attr",
    ):
        assert tracks._get_node_attr(1, "a") == 42

    tracks._set_nodes_attr([1, 2], "b", [7, 8])
    with pytest.warns(
        DeprecationWarning,
        match="_get_nodes_attr deprecated in favor of public method get_nodes_attr",
    ):
        assert tracks._get_nodes_attr([1, 2], "b") == [7, 8]

    # test new functions
    assert tracks.get_node_attr(1, "a", required=True) == 42
    assert tracks.get_nodes_attr([1, 2], "b", required=True) == [7, 8]
    assert tracks.get_nodes_attr([1, 2], "b", required=False) == [7, 8]
    with pytest.raises(KeyError):
        tracks.get_node_attr(1, "not_present", required=True)
    assert tracks.get_node_attr(1, "not_present", required=False) is None
    with pytest.raises(KeyError):
        tracks.get_nodes_attr([1, 2], "not_present", required=True)
    assert all(
        x is None for x in tracks.get_nodes_attr([1, 2], "not_present", required=False)
    )

    # test array attributes
    tracks._set_node_attr(1, "array_attr", np.array([1, 2, 3]))
    tracks._set_nodes_attr((1, 2), "array_attr2", np.array(([1, 2, 3], [4, 5, 6])))


def test_get_set_edge_attr(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    tracks._set_edge_attr((1, 2), "c", 99)
    assert tracks.get_edge_attr((1, 2), "c") == 99
    assert tracks.get_edge_attr((1, 2), "iou", required=True) == 0.0
    tracks._set_edges_attr([(1, 2), (1, 3)], "d", [123, 5])
    assert tracks.get_edges_attr([(1, 2), (1, 3)], "d", required=True) == [123, 5]
    assert tracks.get_edges_attr([(1, 2), (1, 3)], "d", required=False) == [123, 5]
    with pytest.raises(KeyError):
        tracks.get_edge_attr((1, 2), "not_present", required=True)
    assert tracks.get_edge_attr((1, 2), "not_present", required=False) is None
    with pytest.raises(KeyError):
        tracks.get_edges_attr([(1, 2), (1, 3)], "not_present", required=True)
    assert all(
        x is None
        for x in tracks.get_edges_attr(((1, 2), (1, 3)), "not_present", required=False)
    )


def test_set_positions_str(graph_2d_with_computed_features):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    tracks.set_positions((1, 2), [(1, 2), (3, 4)])
    assert np.array_equal(
        tracks.get_positions((1, 2), incl_time=False), np.array([[1, 2], [3, 4]])
    )
    assert np.array_equal(
        tracks.get_positions((1, 2), incl_time=True), np.array([[0, 1, 2], [1, 3, 4]])
    )

    # test invalid node id
    with pytest.raises(KeyError):
        tracks.get_positions(["0"])


def test_set_positions_list(graph_2d_list):
    tracks = Tracks(graph_2d_list, pos_attr=["y", "x"], ndim=3, **track_attrs)
    tracks.set_positions((1, 2), [(1, 2), (3, 4)])
    assert np.array_equal(
        tracks.get_positions((1, 2), incl_time=False), np.array([[1, 2], [3, 4]])
    )
    assert np.array_equal(
        tracks.get_positions((1, 2), incl_time=True), np.array([[0, 1, 2], [1, 3, 4]])
    )


def test_set_node_attributes(graph_2d_with_computed_features, caplog):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    attrs = {"attr_1": 1, "attr_2": ["a", "b", "c", "d", "e", "f"]}
    tracks._set_node_attributes(1, attrs)
    assert tracks.get_node_attr(1, "attr_1") == 1
    assert tracks.get_node_attr(1, "attr_2") == ["a", "b", "c", "d", "e", "f"]
    with caplog.at_level("INFO"):
        tracks._set_node_attributes(7, attrs)
    assert any("Node 7 not found in the graph." in message for message in caplog.messages)


def test_set_edge_attributes(graph_2d_with_computed_features, caplog):
    tracks = Tracks(graph_2d_with_computed_features, ndim=3, **track_attrs)
    attrs = {"attr_1": 1, "attr_2": ["a", "b", "c", "d"]}
    tracks._set_edge_attributes((1, 2), attrs)
    assert tracks.get_edge_attr((1, 2), "attr_1") == 1
    assert tracks.get_edge_attr((1, 2), "attr_2") == ["a", "b", "c", "d"]
    with caplog.at_level("INFO"):
        tracks._set_edge_attributes((4, 6), attrs)
    assert any(
        "Edge (4, 6) not found in the graph." in message for message in caplog.messages
    )


def test_get_pixels_and_set_pixels(graph_2d_with_computed_features, segmentation_2d):
    tracks = Tracks(
        graph_2d_with_computed_features, segmentation_2d, ndim=3, **track_attrs
    )
    pix = tracks.get_pixels(1)
    assert isinstance(pix, tuple)
    tracks.set_pixels(pix, 99)
    assert tracks.segmentation[0, 50, 50] == 99


def test_get_pixels_none(graph_2d_with_computed_features):
    tracks = Tracks(
        graph_2d_with_computed_features, segmentation=None, ndim=3, **track_attrs
    )
    assert tracks.get_pixels([1]) is None


def test_set_pixels_no_segmentation(graph_2d_with_computed_features):
    tracks = Tracks(
        graph_2d_with_computed_features, segmentation=None, ndim=3, **track_attrs
    )
    pix = [(np.array([0]), np.array([10]), np.array([20]))]
    with pytest.raises(ValueError):
        tracks.set_pixels(pix, [1])


def test_compute_ndim_errors():
    g = nx.DiGraph()
    g.add_node(1, time=0, pos=[0, 0, 0])
    # seg ndim = 3, scale ndim = 2, provided ndim = 4 -> mismatch
    seg = np.zeros((2, 2, 2))
    with pytest.raises(ValueError, match="Dimensions from segmentation"):
        Tracks(g, segmentation=seg, scale=[1, 2], ndim=4)

    with pytest.raises(
        ValueError, match="Cannot compute dimensions from segmentation or scale"
    ):
        Tracks(g)



from __future__ import annotations

import warnings
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

from funtracks.actions.add_delete_edge import AddEdge
from funtracks.actions.update_segmentation import UpdateNodeSeg
from funtracks.features import Feature, IoU

from ._compute_ious import _compute_ious
from ._graph_annotator import GraphAnnotator

if TYPE_CHECKING:
    from funtracks.actions import BasicAction
    from funtracks.data_model import Tracks

DEFAULT_IOU_KEY = "iou"


class EdgeAnnotator(GraphAnnotator):
    """Manages edge features computed from segmentations or endpoint positions.

    The possible features include:
    - Intersection over Union (IoU)

    Args:
        tracks (Tracks): The tracks to manage the edge features on
    """

    @classmethod
    def can_annotate(cls, tracks) -> bool:
        """Check if this annotator can annotate the given tracks.

        Requires segmentation data to be present.

        Args:
            tracks: The tracks to check compatibility with

        Returns:
            True if tracks have segmentation, False otherwise
        """
        return tracks.segmentation is not None

    @classmethod
    def get_available_features(cls, ndim: int = 3) -> dict[str, Feature]:
        """Get all features that can be computed by this annotator.

        Returns features with default keys. Custom keys can be specified at
        initialization time.

        Args:
            ndim: Total number of dimensions including time (unused for this annotator,
                kept for API consistency). Defaults to 3.

        Returns:
            Dictionary mapping feature keys to Feature definitions.
        """
        return {DEFAULT_IOU_KEY: IoU()}

    def __init__(self, tracks: Tracks) -> None:
        self.iou_key = DEFAULT_IOU_KEY
        # Build features dict with custom key
        feats = {} if tracks.segmentation is None else {DEFAULT_IOU_KEY: IoU()}
        super().__init__(tracks, feats)

    def compute(self, feature_keys: list[str] | None = None) -> None:
        """Compute the currently included features and add them to the tracks.

        Args:
            feature_keys: Optional list of specific feature keys to compute.
                If None, computes all currently active features. Keys not in
                self.features (not enabled) are ignored.
        """
        # Can only compute features if segmentation is present
        if self.tracks.segmentation is None:
            return

        keys_to_compute = self._filter_feature_keys(feature_keys)
        if not keys_to_compute:
            return

        seg = self.tracks.segmentation
        # TODO: add skip edges
        if self.iou_key in keys_to_compute:
            nodes_by_frame = defaultdict(list)
            for n in self.tracks.nodes():
                nodes_by_frame[self.tracks.get_time(n)].append(n)

            for t in range(seg.shape[0] - 1):
                nodes_in_t = nodes_by_frame[t]
                edges = list(self.tracks.graph.out_edges(nodes_in_t))
                self._iou_update(edges, seg[t], seg[t + 1])

    def _iou_update(
        self,
        edges: list[tuple[int, int]],
        seg_frame: np.ndarray,
        seg_next_frame: np.ndarray,
    ) -> None:
        """Perform the IoU computation and update all feature values for a
        single pair of frames of segmentation data.

        Args:
            edges (list[tuple[int, int]]): A list of edges between two frames
            seg_frame (np.ndarray): A 2D or 3D numpy array representing the seg for the
                starting time of the edges
            seg_next_frame (np.ndarray): A 2D or 3D numpy array representing the seg for
                the ending time of the edges
        """
        ious = _compute_ious(seg_frame, seg_next_frame)  # list of (id1, id2, iou)
        for id1, id2, iou in ious:
            edge = (id1, id2)
            if edge in edges:
                self.tracks._set_edge_attr(edge, self.iou_key, iou)
                edges.remove(edge)

        # anything left has IOU of 0
        for edge in edges:
            self.tracks._set_edge_attr(edge, self.iou_key, 0)

    def update(self, action: BasicAction):
        """Update the edge features based on the action.

        Only responds to AddEdge and UpdateNodeSeg actions that affect edge IoU.

        Args:
            action (BasicAction): The action that triggered this update
        """
        # Only update for actions that change edges or segmentation
        if not isinstance(action, (AddEdge, UpdateNodeSeg)):
            return

        # Can only compute features if segmentation is present
        if self.tracks.segmentation is None:
            return

        if self.iou_key not in self.features:
            return

        # Get edges to update based on action type
        if isinstance(action, AddEdge):
            edges_to_update = [action.edge]
        else:  # UpdateNodeSeg
            # Get all incident edges to the modified node
            node = action.node
            edges_to_update = list(self.tracks.graph.in_edges(node)) + list(
                self.tracks.graph.out_edges(node)
            )

        # Update IoU for each edge
        for edge in edges_to_update:
            source, target = edge
            start_time = self.tracks.get_time(source)
            end_time = self.tracks.get_time(target)
            start_seg = self.tracks.segmentation[start_time]
            end_seg = self.tracks.segmentation[end_time]
            masked_start = np.where(start_seg == source, source, 0)
            masked_end = np.where(end_seg == target, target, 0)
            if np.max(masked_start) == 0 or np.max(masked_end) == 0:
                warnings.warn(
                    f"Cannot find label {source} in frame {start_time} or label {target} "
                    f"in frame {end_time}: updating edge IOU value to 0",
                    stacklevel=2,
                )
                self.tracks._set_edge_attr(edge, self.iou_key, 0)
            else:
                iou_list = _compute_ious(masked_start, masked_end)
                iou = 0 if len(iou_list) == 0 else iou_list[0][2]
                self.tracks._set_edge_attr(edge, self.iou_key, iou)

    def change_key(self, old_key: str, new_key: str) -> None:
        """Rename a feature key in this annotator.

        Overrides base implementation to also update the iou_key instance variable.

        Args:
            old_key: Existing key to rename.
            new_key: New key to replace it with.

        Raises:
            KeyError: If old_key does not exist.
        """
        # Call base implementation to update all_features
        super().change_key(old_key, new_key)

        # Update iou_key if it matches
        if self.iou_key == old_key:
            self.iou_key = new_key



"""Tests for validation functions in funtracks.import_export._validation."""

import pytest

from funtracks.import_export._validation import (
    validate_edge_name_map,
    validate_feature_key_collisions,
    validate_node_name_map,
)


class TestValidateNodeNameMap:
    """Test validate_node_name_map helper function."""

    def test_valid_node_name_map(self):
        """Test that a valid node name_map passes validation."""
        name_map = {"time": "t", "x": "x_coord", "y": "y_coord"}
        importable_props = ["t", "x_coord", "y_coord", "area"]
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3  # 2D + time

        # Should not raise
        validate_node_name_map(
            name_map, importable_props, required_features, position_attr, ndim
        )

    def test_missing_required_feature(self):
        """Test that missing required features raise ValueError."""
        name_map = {"x": "x_coord", "y": "y_coord"}  # Missing "time"
        importable_props = ["t", "x_coord", "y_coord"]
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3

        with pytest.raises(ValueError, match="None values"):
            validate_node_name_map(
                name_map, importable_props, required_features, position_attr, ndim
            )

    def test_missing_position_attr(self):
        """Test that missing position attributes raise ValueError."""
        name_map = {"time": "t", "x": "x_coord"}  # Missing "y"
        importable_props = ["t", "x_coord", "y_coord"]
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3

        with pytest.raises(ValueError, match="None values"):
            validate_node_name_map(
                name_map, importable_props, required_features, position_attr, ndim
            )

    def test_none_value_in_required_field(self):
        """Test that None values in required fields raise ValueError."""
        name_map = {"time": "t", "x": None, "y": "y_coord"}
        importable_props = ["t", "y_coord"]
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3

        with pytest.raises(ValueError, match="cannot contain None values"):
            validate_node_name_map(
                name_map, importable_props, required_features, position_attr, ndim
            )

    def test_duplicate_values_in_required_fields(self):
        """Test that duplicate values in required fields are allowed.

        Multiple standard keys can map to the same source property.
        This is useful for cases like seg_id mapping to the same column as id.
        """
        name_map = {"time": "t", "x": "coord", "y": "coord"}  # Duplicate "coord"
        importable_props = ["t", "coord"]
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3

        # Should not raise - duplicates are allowed
        validate_node_name_map(
            name_map, importable_props, required_features, position_attr, ndim
        )

    def test_nonexistent_property(self):
        """Test that mapping to non-existent properties raises ValueError."""
        name_map = {"time": "t", "x": "x_coord", "y": "y_coord"}
        importable_props = ["t", "x_coord"]  # "y_coord" doesn't exist
        required_features = ["time"]
        position_attr = ["z", "y", "x"]
        ndim = 3

        with pytest.raises(ValueError, match="non-existent properties"):
            validate_node_name_map(
                name_map, importable_props, required_features, position_attr, ndim
            )


class TestValidateEdgeNameMap:
    """Test validate_edge_name_map helper function."""

    def test_valid_edge_name_map(self):
        """Test that a valid edge name_map passes validation."""
        edge_name_map = {"iou": "overlap", "distance": "dist"}
        importable_props = ["overlap", "dist", "weight"]

        # Should not raise
        validate_edge_name_map(edge_name_map, importable_props)

    def test_nonexistent_edge_property(self):
        """Test that mapping to non-existent edge properties raises ValueError."""
        edge_name_map = {"iou": "overlap", "distance": "dist"}
        importable_props = ["overlap"]  # "dist" doesn't exist

        with pytest.raises(ValueError, match="non-existent properties"):
            validate_edge_name_map(edge_name_map, importable_props)

    def test_empty_importable_props(self):
        """Test that empty importable_props list doesn't raise."""
        edge_name_map = {"iou": "overlap"}
        importable_props = []

        # Should not raise when importable_props is empty
        validate_edge_name_map(edge_name_map, importable_props)


class TestValidateFeatureKeyCollisions:
    """Test validate_feature_key_collisions helper function."""

    def test_no_collision(self):
        """Test that non-overlapping keys pass validation."""
        name_map = {"time": "t", "x": "x", "y": "y", "area": "area"}
        edge_name_map = {"iou": "iou", "distance": "distance"}

        # Should not raise
        validate_feature_key_collisions(name_map, edge_name_map)

    def test_single_collision(self):
        """Test that a single colliding key raises ValueError."""
        name_map = {"time": "t", "x": "x", "y": "y", "iou": "node_iou"}
        edge_name_map = {"iou": "edge_iou", "distance": "distance"}

        with pytest.raises(ValueError, match="Feature keys cannot be shared"):
            validate_feature_key_collisions(name_map, edge_name_map)

        with pytest.raises(ValueError, match="iou"):
            validate_feature_key_collisions(name_map, edge_name_map)

    def test_multiple_collisions(self):
        """Test that multiple colliding keys are all reported."""
        name_map = {
            "time": "t",
            "x": "x",
            "y": "y",
            "iou": "node_iou",
            "weight": "node_weight",
        }
        edge_name_map = {"iou": "edge_iou", "weight": "edge_weight", "distance": "dist"}

        with pytest.raises(ValueError, match="Feature keys cannot be shared"):
            validate_feature_key_collisions(name_map, edge_name_map)

        with pytest.raises(ValueError, match="iou"):
            validate_feature_key_collisions(name_map, edge_name_map)

        with pytest.raises(ValueError, match="weight"):
            validate_feature_key_collisions(name_map, edge_name_map)

    def test_none_edge_name_map(self):
        """Test that None edge_name_map doesn't raise."""
        name_map = {"time": "t", "x": "x", "y": "y", "iou": "iou"}
        edge_name_map = None

        # Should not raise
        validate_feature_key_collisions(name_map, edge_name_map)



"""Deprecated: Import from funtracks.import_export instead.

This module is deprecated and will be removed in funtracks v2.0.
Use: from funtracks.import_export import import_from_geff

.. deprecated:: 1.0
    This module location is deprecated. Import from the main package instead:
    ``from funtracks.import_export import import_from_geff``
"""

from __future__ import annotations

import warnings

from .geff._import import import_from_geff, import_graph_from_geff

warnings.warn(
    "Importing from funtracks.import_export.import_from_geff is deprecated. "
    "Use 'from funtracks.import_export import import_from_geff' instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["import_from_geff", "import_graph_from_geff"]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._base import Action


class ActionHistory:
    """An action history implementing the ideas from this blog:
    https://github.com/zaboople/klonk/blob/master/TheGURQ.md
    Essentially, if you go back and change something after undo-ing, you can always get
    back to every state if you undo far enough (instead of throwing out
    the undone actions)
    """

    def __init__(self) -> None:
        self.undo_stack: list[Action] = []  # list of actions that can be undone
        self.redo_stack: list[Action] = []  # list of actions that can be redone

    @property
    def _undo_pointer(self):
        return len(self.undo_stack) - len(self.redo_stack) - 1

    def add_new_action(self, action: Action) -> None:
        """Add a newly performed action to the history.
        Args:
            action (Action): The new action to be added to the history.
        """
        if len(self.redo_stack) > 0:
            # add all the redo stuff to the undo stack, so that both the original and
            # inverse are on the stack
            self.undo_stack.extend(self.redo_stack)
            self.redo_stack = []
        self.undo_stack.append(action)

    def undo(self) -> bool:
        """Undo the last performed action

        Returns:
            bool: True if an action was undone, and False
            if there was no previous action to undo.
        """
        if self._undo_pointer < 0:
            return False
        else:
            action = self.undo_stack[self._undo_pointer]
            inverse = action.inverse()
            self.redo_stack.append(inverse)
            return True

    def redo(self) -> bool:
        """Redo the last undone action

        Returns:
            bool: True if an action was redone, and False
            if there was no undone action to redo.
        """
        if len(self.redo_stack) == 0:
            return False
        else:
            action = self.redo_stack.pop(-1)
            # apply the inverse but don't save it
            # (the original is already on the undo stack)
            action.inverse()
            return True



from __future__ import annotations

import json
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import networkx as nx
import numpy as np

from funtracks.features import FeatureDict

if TYPE_CHECKING:
    from ..data_model import SolutionTracks, Tracks

GRAPH_FILE = "graph.json"
SEG_FILE = "seg.npy"
ATTRS_FILE = "attrs.json"


def save_tracks(tracks: Tracks, directory: Path) -> None:
    """Save the tracks to the given directory.
    Currently, saves the graph as a json file in networkx node link data format,
    saves the segmentation as a numpy npz file, and saves the time and position
    attributes and scale information in an attributes json file.
    Will make the directory if it doesn't exist.

    Args:
        tracks (Tracks): the tracks to save
        directory (Path): The directory to save the tracks in.
    """
    directory.mkdir(exist_ok=True, parents=True)
    _save_graph(tracks, directory)
    _save_seg(tracks, directory)
    _save_attrs(tracks, directory)


def _save_graph(tracks: Tracks, directory: Path) -> None:
    """Save the graph to file. Currently uses networkx node link data
    format (and saves it as json).

    Args:
        tracks (Tracks): the tracks to save the graph of
        directory (Path): The directory in which to save the graph file.
    """
    graph_file = directory / GRAPH_FILE
    graph_data = nx.node_link_data(tracks.graph, edges="links")

    def convert_np_types(data):
        """Recursively convert numpy types to native Python types."""

        if isinstance(data, dict):
            return {key: convert_np_types(value) for key, value in data.items()}
        elif isinstance(data, list):
            return [convert_np_types(item) for item in data]
        elif isinstance(data, np.ndarray):
            return data.tolist()  # Convert numpy arrays to Python lists
        elif isinstance(data, np.integer):
            return int(data)  # Convert numpy integers to Python int
        elif isinstance(data, np.floating):
            return float(data)  # Convert numpy floats to Python float
        else:
            return data  # Return the data as-is if it's already a native Python type

    graph_data = convert_np_types(graph_data)
    with open(graph_file, "w") as f:
        json.dump(graph_data, f)


def _save_seg(tracks: Tracks, directory: Path) -> None:
    """Save a segmentation as a numpy array using np.save. In the future,
    could be changed to use zarr or other file types.

    Args:
        tracks (Tracks): the tracks to save the segmentation of
        directory (Path): The directory in which to save the segmentation
    """
    if tracks.segmentation is not None:
        out_path = directory / SEG_FILE
        np.save(out_path, tracks.segmentation)


def _save_attrs(tracks: Tracks, directory: Path) -> None:
    """Save the and scale, ndim, and features in a json file in the given directory.

    Args:
        tracks (Tracks): the tracks to save the attributes of
        directory (Path):  The directory in which to save the attributes
    """
    out_path = directory / ATTRS_FILE
    attrs_dict = {
        "scale": tracks.scale
        if not isinstance(tracks.scale, np.ndarray)
        else tracks.scale.tolist(),
        "ndim": tracks.ndim,
        "features": tracks.features.dump_json(),
    }
    with open(out_path, "w") as f:
        json.dump(attrs_dict, f)


def load_tracks(
    directory: Path, seg_required: bool = False, solution: bool = False
) -> Tracks | SolutionTracks:
    """Load a Tracks object from the given directory. Looks for files
    in the format generated by Tracks.save.

    Args:
        directory (Path): The directory containing tracks to load
        seg_required (bool, optional): If true, raises a FileNotFoundError if the
            segmentation file is not present in the directory. Defaults to False.
        solution (bool, optional): If true, returns a SolutionTracks object, otherwise
            returns a normal Tracks object. Defaults to False.

    Returns:
        Tracks: A tracks object loaded from the given directory
    """
    graph_file = directory / GRAPH_FILE
    graph = _load_graph(graph_file)

    seg_file = directory / SEG_FILE
    seg = _load_seg(seg_file, seg_required=seg_required)

    attrs_file = directory / ATTRS_FILE
    attrs = _load_attrs(attrs_file)

    # filtering the warnings because the default values of time_attr and pos_attr are
    # not None. Therefore, new style Tracks attrs that have features instead of
    # pos_attr and time_attr will always trigger the warning. Updating default values
    # is breaking, and manually setting the attrs to None if features is present will
    # break if the attrs are changed/removed in the future. Can remove in v2.0.
    # Import at runtime to avoid circular dependency
    from ..data_model import SolutionTracks, Tracks

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="Provided both FeatureDict and pos_attr or time_attr"
        )
        tracks: Tracks
        if solution:
            tracks = SolutionTracks(graph, seg, **attrs)
        else:
            tracks = Tracks(graph, seg, **attrs)
    return tracks


def _load_graph(graph_file: Path) -> nx.DiGraph:
    """Load the graph from the given json file. Expects networkx node_link_graph
    formatted json.

    Args:
        graph_file (Path): The json file to load into a networkx graph

    Raises:
        FileNotFoundError: If the file does not exist

    Returns:
        nx.DiGraph: A networkx graph loaded from the file.
    """
    if graph_file.is_file():
        with open(graph_file) as f:
            json_graph = json.load(f)
        return nx.node_link_graph(json_graph, directed=True, edges="links")
    else:
        raise FileNotFoundError(f"No graph at {graph_file}")


def _load_seg(seg_file: Path, seg_required: bool = False) -> np.ndarray | None:
    """Load a segmentation from a file. If the file doesn't exist, either return
    None or raise a FileNotFoundError depending on the seg_required flag.

    Args:
        seg_file (Path): The npz file to load.
        seg_required (bool, optional): If true, raise a FileNotFoundError if the
            segmentation is not present. Defaults to False.

    Returns:
        np.ndarray | None: The segmentation array, or None if it wasn't present and
            seg_required was False.
    """
    if seg_file.is_file():
        return np.load(seg_file)
    elif seg_required:
        raise FileNotFoundError(f"No segmentation at {seg_file}")
    else:
        return None


def _load_attrs(attrs_file: Path) -> dict:
    if attrs_file.is_file():
        with open(attrs_file) as f:
            json_dict = json.load(f)
        if "features" in json_dict:
            json_dict["features"] = FeatureDict.from_json(json_dict["features"])
        return json_dict

    else:
        raise FileNotFoundError(f"No attributes at {attrs_file}")


def delete_tracks(directory: Path | str):
    """Delete the tracks in the given directory. Also deletes the directory.

    Args:
        directory (Path | str): Path to a directory containing tracks
    """
    if not isinstance(directory, Path):
        directory = Path(directory)
    (directory / GRAPH_FILE).unlink()
    (directory / SEG_FILE).unlink(missing_ok=True)
    (directory / ATTRS_FILE).unlink()
    directory.rmdir()



import pytest

from funtracks.actions import (
    Action,
)


def test_initialize_base_class(get_tracks):
    tracks = get_tracks(ndim=3, with_seg=True, is_solution=True)
    action = Action(tracks)
    with pytest.raises(NotImplementedError):
        action.inverse()



import numpy as np


def _compute_ious(frame1: np.ndarray, frame2: np.ndarray) -> list[tuple[int, int, float]]:
    """Compute label IOUs between two label arrays of the same shape. Ignores background
    (label 0).

    Args:
        frame1 (np.ndarray): Array with integer labels
        frame2 (np.ndarray): Array with integer labels

    Returns:
        list[tuple[int, int, float]]: List of tuples of label in frame 1, label in
            frame 2, and iou values. Labels that have no overlap are not included.
    """
    frame1 = frame1.flatten()
    frame2 = frame2.flatten()
    # get indices where both are not zero (ignore background)
    # this speeds up computation significantly
    non_zero_indices = np.logical_and(frame1, frame2)
    flattened_stacked = np.array([frame1[non_zero_indices], frame2[non_zero_indices]])

    values, counts = np.unique(flattened_stacked, axis=1, return_counts=True)
    frame1_values, frame1_counts = np.unique(frame1, return_counts=True)
    frame1_label_sizes = dict(zip(frame1_values, frame1_counts, strict=True))
    frame2_values, frame2_counts = np.unique(frame2, return_counts=True)
    frame2_label_sizes = dict(zip(frame2_values, frame2_counts, strict=True))
    ious: list[tuple[int, int, float]] = []
    for index in range(values.shape[1]):
        pair = values[:, index]
        intersection = counts[index]
        id1, id2 = pair
        union = frame1_label_sizes[id1] + frame2_label_sizes[id2] - intersection
        ious.append((id1, id2, intersection / union))
    return ious



from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

import networkx as nx
from geff._typing import InMemoryGeff
from geff.validate.graph import (
    validate_no_repeated_edges,
    validate_no_self_edges,
    validate_nodes_for_edges,
    validate_unique_node_ids,
)
from geff.validate.segmentation import has_seg_ids_at_coords
from geff.validate.tracks import validate_lineages, validate_tracklets

if TYPE_CHECKING:
    import dask.array as da

# Constants from import_from_geff
SEG_KEY = "seg_id"


def validate_graph_seg_match(
    graph: nx.DiGraph,
    segmentation: da.Array,
    scale: list[float],
    position_attr: list[str],
) -> bool:
    """Validate if the graph matches the provided segmentation data.

    Checks if the seg_id value of the last node matches the pixel value at the
    (scaled) node coordinates. Returns a boolean indicating whether relabeling
    of the segmentation to match node id values is required.

    Args:
        graph: NetworkX graph with standard keys
        segmentation: Segmentation data (dask array)
        scale: Scaling information (pixel to world coordinates)
        position_attr: Position keys (e.g., ["y", "x"] or ["z", "y", "x"])

    Returns:
        bool: True if relabeling from seg_id to node_id is required.
    """
    # Check segmentation dimensions match graph dimensionality
    ndim = len(position_attr) + 1  # +1 for time
    if len(segmentation.shape) != ndim:
        raise ValueError(
            f"Segmentation has {len(segmentation.shape)} dimensions but graph has "
            f"{ndim} dimensions (time + {len(position_attr)} spatial dims)"
        )

    # Get the last node for validation
    node_ids = list(graph.nodes())
    if not node_ids:
        raise ValueError("Graph has no nodes")

    last_node_id = node_ids[-1]
    last_node_data = graph.nodes[last_node_id]

    # Check if seg_id exists; if not, assume it matches node_id
    seg_id = last_node_data.get(SEG_KEY, last_node_id)

    # Get the coordinates for the last node (using standard keys)
    coord = [int(last_node_data["time"])]
    if "z" in position_attr:
        coord.append(last_node_data["z"])
    coord.append(last_node_data["y"])
    coord.append(last_node_data["x"])

    # Check bounds
    for i, (c, s) in enumerate(zip(coord, segmentation.shape, strict=False)):
        pixel_coord = int(c / scale[i])
        if not (0 <= pixel_coord < s):
            raise ValueError(
                f"Coordinate {i} ({c}) is out of bounds for segmentation shape {s} "
                f"(pixel coord: {pixel_coord})"
            )

    # Check if the segmentation pixel value at the coordinates matches the seg id
    seg_id_at_coord, errors = has_seg_ids_at_coords(
        segmentation, [coord], [seg_id], tuple(1 / s for s in scale)
    )
    if not seg_id_at_coord:
        error_msg = "Error testing seg id:\n" + "\n".join(f"- {e}" for e in errors)
        raise ValueError(error_msg)

    # TODO: The relabeling check (seg_id != node_id) is duplicated in
    # TracksBuilder.handle_segmentation. Consider deduplicating by either:
    # 1. Using this return value in the caller, or
    # 2. Removing the return value and making this purely a validation function
    # Return True if relabeling is needed (seg_id != node_id)
    return last_node_id != seg_id


def validate_node_name_map(
    name_map: dict[str, str],
    importable_node_props: list[str],
    required_features: list[str],
    position_attr: list[str],
    ndim: int | None,
) -> None:
    """Validate node name_map contains all required mappings.

    Checks:
    - No None values in required mappings
    - No duplicate values in required mappings
    - All required_features are mapped
    - All position_attr are mapped (based on ndim)
    - All mapped properties exist in importable_node_props

    Args:
        name_map: Mapping from standard keys to source property names
        importable_node_props: List of property names available in the source
        required_features: List of required feature names (e.g., ["time"])
        position_attr: List of position attributes (e.g., ["z", "y", "x"])
        ndim: Number of dimensions (3 for 2D+time, 4 for 3D+time), or None to
            accept either 2D or 3D

    Raises:
        ValueError: If validation fails
    """
    # Build list of required position attributes
    # When ndim is None, require y and x at minimum; z is optional
    if ndim is None:
        ndim = 3
    required_pos_attrs = position_attr[-(ndim - 1) :]

    required_fields = required_features + required_pos_attrs

    # Check for None values in required fields
    none_mappings = [key for key in required_fields if name_map.get(key) is None]
    if none_mappings:
        raise ValueError(
            f"The name_map cannot contain None values. "
            f"Fields with None values: {none_mappings}"
        )

    missing_features = []

    # Check required features
    for feature in required_fields:
        if feature not in name_map:
            missing_features.append(feature)

    if missing_features:
        raise ValueError(
            f"name_map is missing required mappings: {missing_features}. "
            f"Required features: {required_features}. "
            f"Required position attrs: {required_pos_attrs}"
        )

    # Fail if mapped properties don't exist in importable_node_props
    if importable_node_props:
        invalid_mappings = []
        for std_key, source_prop in name_map.items():
            if source_prop not in importable_node_props:
                invalid_mappings.append(f"{std_key} -> '{source_prop}'")

        if invalid_mappings:
            raise ValueError(
                f"name_map contains mappings to non-existent properties: "
                f"{invalid_mappings}. "
                f"Importable node properties: {importable_node_props}"
            )


def validate_edge_name_map(
    edge_name_map: dict[str, str],
    importable_edge_props: list[str],
) -> None:
    """Validate edge name_map mappings exist in source.

    Checks:
    - All mapped edge properties exist in importable_edge_props

    Args:
        edge_name_map: Mapping from standard keys to edge property names
        importable_edge_props: List of edge property names available in the source

    Raises:
        ValueError: If validation fails
    """
    if not importable_edge_props:
        return

    invalid_mappings = []
    for std_key, source_prop in edge_name_map.items():
        if source_prop not in importable_edge_props:
            invalid_mappings.append(f"{std_key} -> '{source_prop}'")

    if invalid_mappings:
        raise ValueError(
            f"edge_name_map contains mappings to non-existent properties: "
            f"{invalid_mappings}. "
            f"Importable edge properties: {importable_edge_props}"
        )


def validate_feature_key_collisions(
    name_map: dict[str, str],
    edge_name_map: dict[str, str] | None,
) -> None:
    """Validate that node and edge feature keys don't overlap.

    Feature keys must be unique across both node and edge features because
    they share the same namespace in FeatureDict.

    Args:
        name_map: Mapping from standard keys to node property names
        edge_name_map: Optional mapping from standard keys to edge property names

    Raises:
        ValueError: If any keys appear in both name_map and edge_name_map
    """
    if edge_name_map is None:
        return

    # Get the standard keys (not the source property names) from both maps
    node_keys = set(name_map.keys())
    edge_keys = set(edge_name_map.keys())

    # Find overlapping keys
    colliding_keys = node_keys & edge_keys

    if colliding_keys:
        raise ValueError(
            f"Feature keys cannot be shared between nodes and edges. "
            f"Colliding keys: {sorted(colliding_keys)}. "
            f"Please use unique keys for node and edge features."
        )


def validate_in_memory_geff(in_memory_geff: InMemoryGeff) -> None:
    """Validate the loaded InMemoryGeff data.

    Validates graph structure (required - raises on failure):
    - validate_unique_node_ids: No duplicate node IDs
    - validate_nodes_for_edges: All edges reference existing nodes
    - validate_no_self_edges: No self-loops
    - validate_no_repeated_edges: No duplicate edges

    Validates optional properties (warns and removes if invalid):
    - validate_tracklets: track_id must form valid tracklets
    - validate_lineages: lineage_id must form valid lineages

    Args:
        in_memory_geff: InMemoryGeff data structure to validate

    Raises:
        ValueError: If required validation (graph structure) fails
    """
    node_ids = in_memory_geff["node_ids"]
    edge_ids = in_memory_geff["edge_ids"]
    node_props = in_memory_geff["node_props"]

    # Validate graph structure (required - always fails if invalid)
    valid, nonunique_nodes = validate_unique_node_ids(node_ids)
    if not valid:
        raise ValueError(f"Some node ids are not unique:\n{nonunique_nodes}")

    valid, invalid_edges = validate_nodes_for_edges(node_ids, edge_ids)
    if not valid:
        raise ValueError(f"Some edges are missing nodes:\n{invalid_edges}")

    valid, invalid_edges = validate_no_self_edges(edge_ids)
    if not valid:
        raise ValueError(f"Self edges found in data:\n{invalid_edges}")

    valid, invalid_edges = validate_no_repeated_edges(edge_ids)
    if not valid:
        raise ValueError(f"Repeated edges found in data:\n{invalid_edges}")

    # Validate tracklet_id if present (optional - remove if invalid)
    if "track_id" in node_props:
        tracklet_ids = node_props["track_id"]["values"]
        valid, errors = validate_tracklets(node_ids, edge_ids, tracklet_ids)
        if not valid:
            warn(
                f"track_id validation failed:\n{chr(10).join(errors)}\n"
                "Removing track_id from data.",
                stacklevel=2,
            )
            del node_props["track_id"]

    # Validate lineage_id if present (optional - remove if invalid)
    if "lineage_id" in node_props:
        lineage_ids = node_props["lineage_id"]["values"]
        valid, errors = validate_lineages(node_ids, edge_ids, lineage_ids)
        if not valid:
            warn(
                f"lineage_id validation failed:\n{chr(10).join(errors)}\n"
                "Removing lineage_id from data.",
                stacklevel=2,
            )
            del node_props["lineage_id"]



from __future__ import annotations

import itertools
from typing import (
    TYPE_CHECKING,
    Literal,
)

import geff
import geff_spec
import networkx as nx
import numpy as np
from geff_spec import GeffMetadata

from funtracks.utils import remove_tilde, setup_zarr_array, setup_zarr_group

from .._utils import filter_graph_with_ancestors

if TYPE_CHECKING:
    from pathlib import Path

    from funtracks.data_model.tracks import Tracks


def export_to_geff(
    tracks: Tracks,
    directory: Path,
    overwrite: bool = False,
    node_ids: set[int] | None = None,
    zarr_format: Literal[2, 3] = 2,
):
    """Export the Tracks nxgraph to geff.

    Args:
        tracks (Tracks): Tracks object containing a graph to save.
        directory (Path): Destination directory for saving the Zarr.
        overwrite (bool): If True, allows writing into a non-empty directory.
        node_ids (set[int], optional): A set of nodes that should be saved. If
            provided, a valid graph will be constructed that also includes the ancestors
            of the given nodes. All other nodes will NOT be saved.
        zarr_format (Literal[2, 3]): Zarr format version to use. Defaults to 2
            for maximum compatibility.
    """
    directory = remove_tilde(directory)
    directory = directory.resolve(strict=False)

    if node_ids is not None:
        nodes_to_keep = filter_graph_with_ancestors(
            tracks.graph, node_ids
        )  # include the ancestors to make sure the graph is valid and has no missing
        # parent nodes.

    # Create directory as a zarr group
    mode: Literal["w", "w-"] = "w" if overwrite else "w-"
    setup_zarr_group(directory, zarr_format=zarr_format, mode=mode)

    # update the graph to split the position into separate attrs, if they are currently
    # together in a list
    graph, axis_names = split_position_attr(tracks)
    if axis_names is None:
        axis_names = []
    axis_names.insert(0, tracks.features.time_key)
    if axis_names is not None:
        axis_types = (
            ["time", "space", "space"]
            if tracks.ndim == 3
            else ["time", "space", "space", "space"]
        )
    else:
        axis_types = None
    if tracks.scale is None:
        tracks.scale = (1.0,) * tracks.ndim

    metadata = GeffMetadata(
        geff_version=geff_spec.__version__,
        directed=isinstance(graph, nx.DiGraph),
        node_props_metadata={},
        edge_props_metadata={},
    )

    # Save segmentation if present

    # TODO: helper function in _export_segmentation file
    if tracks.segmentation is not None:
        seg_path = directory / "segmentation"

        seg_data = tracks.segmentation
        shape = seg_data.shape
        dtype = seg_data.dtype

        # TODO: this probably isn't a good chunk size - time should be 1?
        # TODO: export to tiffs
        chunk_size: tuple[int, ...] = (64, 64, 64)
        chunk_size = tuple(list(chunk_size) + [1] * (len(shape) - len(chunk_size)))
        chunk_size = chunk_size[: len(shape)]

        z = setup_zarr_array(
            seg_path,
            zarr_format=zarr_format,
            shape=shape,
            dtype=dtype,
            chunks=chunk_size,
        )

        if node_ids is not None:
            nodes_to_keep = np.asarray(nodes_to_keep)

            # to avoid having to copy the segmentation array entirely, loop over chunks,
            # and mask out the nodes that should be kept.
            chunk_ranges = [
                range(0, dim, chunk) for dim, chunk in zip(shape, chunk_size, strict=True)
            ]

            for starts in itertools.product(*chunk_ranges):
                slices = tuple(
                    slice(start, min(start + chunk, dim))
                    for start, chunk, dim in zip(starts, chunk_size, shape, strict=True)
                )

                block = seg_data[slices]
                mask = np.isin(block, nodes_to_keep)
                filtered = np.where(mask, block, 0)
                z[slices] = filtered

        else:
            z[:] = seg_data

        metadata.related_objects = [
            {
                "path": "../segmentation",
                "type": "labels",
                "label_prop": "seg_id",
            }
        ]

    # Filter the graph if node_ids is provided
    if node_ids is not None:
        graph = graph.subgraph(nodes_to_keep).copy()

    # Save the graph in a 'tracks' folder
    tracks_path = directory / "tracks"
    geff.write(
        graph=graph,
        store=tracks_path,
        metadata=metadata,
        axis_names=axis_names,
        axis_types=axis_types,
        axis_scales=tracks.scale,
        overwrite=overwrite,
        zarr_format=zarr_format,
    )


def split_position_attr(tracks: Tracks) -> tuple[nx.DiGraph, list[str] | None]:
    # TODO: this exists in unsqueeze in geff somehow?
    """Spread the spatial coordinates to separate node attrs in order to export to geff
    format.

    Args:
        tracks (funtracks.data_model.Tracks): tracks object holding the graph to be
          converted.

    Returns:
        tuple[nx.DiGraph, list[str]]: graph with a separate positional attribute for each
            coordinate, and the axis names used to store the separate attributes

    """
    pos_key = tracks.features.position_key

    if isinstance(pos_key, str):
        # Position is stored as a single attribute, need to split
        new_graph = tracks.graph.copy()
        new_keys = ["y", "x"]
        if tracks.ndim == 4:
            new_keys.insert(0, "z")
        for _, attrs in new_graph.nodes(data=True):
            pos = attrs.pop(pos_key)
            for i in range(len(new_keys)):
                attrs[new_keys[i]] = pos[i]

        return new_graph, new_keys
    elif pos_key is not None:
        # Position is already split into separate attributes
        return tracks.graph, list(pos_key)
    else:
        return tracks.graph, None



from funtracks.features import (
    Area,
    Circularity,
    EllipsoidAxes,
    Perimeter,
    Position,
    Time,
)


def test_time_feature():
    """Test that Time() returns a valid Feature TypedDict"""
    feat = Time()
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "int"
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Time"
    assert feat["required"] is True
    assert feat["default_value"] is None


def test_position_feature():
    """Test that Position() returns a valid Feature TypedDict"""
    feat = Position(axes=["y", "x"])
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "float"
    assert feat["num_values"] == 2
    assert feat["display_name"] == ["y", "x"]
    assert feat["required"] is True
    assert feat["default_value"] is None


def test_area_feature():
    """Test that Area() returns a valid Feature TypedDict"""
    feat = Area(ndim=3)
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "float"
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Area"
    assert feat["required"] is True
    assert feat["default_value"] is None

    feat = Area(ndim=4)
    assert feat["display_name"] == "Volume"


def test_ellipsoid_axes_feature():
    """Test that EllipsoidAxes() returns a valid Feature TypedDict"""
    feat = EllipsoidAxes(ndim=3)
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "float"
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Ellipse axis radii"

    feat = EllipsoidAxes(ndim=4)
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Ellipsoid axis radii"


def test_circularity_feature():
    """Test that Circularity() returns a valid Feature TypedDict"""
    feat = Circularity(ndim=3)
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "float"
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Circularity"

    feat = Circularity(ndim=4)
    assert feat["display_name"] == "Sphericity"


def test_perimeter_feature():
    """Test that Perimeter() returns a valid Feature TypedDict"""
    feat = Perimeter(ndim=3)
    assert feat["feature_type"] == "node"
    assert feat["value_type"] == "float"
    assert feat["num_values"] == 1
    assert feat["display_name"] == "Perimeter"

    feat = Perimeter(ndim=4)
    assert feat["display_name"] == "Surface Area"


def test_feature_as_dict():
    """Test that Features are valid dicts"""
    feat = Time()
    assert isinstance(feat, dict)
    assert "feature_type" in feat
    assert "value_type" in feat

    # Can convert to regular dict
    regular_dict = dict(feat)
    assert regular_dict == feat



"""Tests for _import_segmentation module."""

import networkx as nx
import numpy as np
import tifffile

from funtracks.import_export._import_segmentation import (
    load_segmentation,
    relabel_segmentation,
)


class TestLoadSegmentation:
    """Tests for load_segmentation function."""

    def test_load_from_path(self, tmp_path):
        """Test loading segmentation from a tif file."""
        # Create test segmentation
        seg = np.array([[[1, 0], [0, 2]], [[3, 0], [0, 4]]], dtype=np.uint16)
        seg_path = tmp_path / "test_seg.tif"
        tifffile.imwrite(seg_path, seg)

        # Load and verify
        result = load_segmentation(seg_path)
        np.testing.assert_array_equal(result.compute(), seg)

    def test_load_from_array(self):
        """Test wrapping a numpy array in dask."""
        seg = np.array([[[1, 0], [0, 2]], [[3, 0], [0, 4]]], dtype=np.uint16)

        result = load_segmentation(seg)

        # Should be a dask array
        assert hasattr(result, "compute")
        np.testing.assert_array_equal(result.compute(), seg)


class TestRelabelSegmentation:
    """Tests for relabel_segmentation function."""

    def test_basic_relabeling(self):
        """Test basic seg_id to node_id relabeling."""
        # Create segmentation with seg_ids 10, 20
        seg = np.zeros((2, 5, 5), dtype=np.uint16)
        seg[0, 1, 1] = 10  # seg_id 10 at t=0
        seg[1, 2, 2] = 20  # seg_id 20 at t=1

        # Create graph with node_ids 1, 2
        graph = nx.DiGraph()
        graph.add_node(1)
        graph.add_node(2)

        node_ids = np.array([1, 2])
        seg_ids = np.array([10, 20])
        time_values = np.array([0, 1])

        result = relabel_segmentation(seg, graph, node_ids, seg_ids, time_values)

        # seg_id 10 -> node_id 1, seg_id 20 -> node_id 2
        assert result[0, 1, 1] == 1
        assert result[1, 2, 2] == 2
        # Background should remain 0
        assert result[0, 0, 0] == 0

    def test_relabeling_with_node_id_zero(self):
        """Test that node_id 0 is handled by offsetting all IDs."""
        # Create segmentation with seg_ids 10, 20
        seg = np.zeros((2, 5, 5), dtype=np.uint16)
        seg[0, 1, 1] = 10  # seg_id 10 at t=0
        seg[1, 2, 2] = 20  # seg_id 20 at t=1

        # Create graph with node_ids 0, 1 (includes 0!)
        graph = nx.DiGraph()
        graph.add_node(0)
        graph.add_node(1)

        node_ids = np.array([0, 1])
        seg_ids = np.array([10, 20])
        time_values = np.array([0, 1])

        result = relabel_segmentation(seg, graph, node_ids, seg_ids, time_values)

        # node_ids should be offset by 1: 0->1, 1->2
        # seg_id 10 -> node_id 1 (was 0), seg_id 20 -> node_id 2 (was 1)
        assert result[0, 1, 1] == 1
        assert result[1, 2, 2] == 2

        # Graph should also be relabeled
        assert 1 in graph.nodes()
        assert 2 in graph.nodes()
        assert 0 not in graph.nodes()

    def test_no_relabeling_needed_same_ids(self):
        """Test when seg_ids equal node_ids (relabeling still applies mapping)."""
        # Create segmentation with seg_ids 1, 2
        seg = np.zeros((2, 5, 5), dtype=np.uint16)
        seg[0, 1, 1] = 1
        seg[1, 2, 2] = 2

        graph = nx.DiGraph()
        graph.add_node(1)
        graph.add_node(2)

        node_ids = np.array([1, 2])
        seg_ids = np.array([1, 2])  # Same as node_ids
        time_values = np.array([0, 1])

        result = relabel_segmentation(seg, graph, node_ids, seg_ids, time_values)

        # Should still produce valid output (identity mapping)
        assert result[0, 1, 1] == 1
        assert result[1, 2, 2] == 2

    def test_multiple_nodes_same_timepoint(self):
        """Test relabeling with multiple nodes at the same timepoint."""
        # Create segmentation with seg_ids 10, 20, 30 at t=0
        seg = np.zeros((1, 5, 5), dtype=np.uint16)
        seg[0, 1, 1] = 10
        seg[0, 2, 2] = 20
        seg[0, 3, 3] = 30

        graph = nx.DiGraph()
        graph.add_nodes_from([1, 2, 3])

        node_ids = np.array([1, 2, 3])
        seg_ids = np.array([10, 20, 30])
        time_values = np.array([0, 0, 0])

        result = relabel_segmentation(seg, graph, node_ids, seg_ids, time_values)

        assert result[0, 1, 1] == 1
        assert result[0, 2, 2] == 2
        assert result[0, 3, 3] == 3



from __future__ import annotations

from typing import TYPE_CHECKING

from ._base import BasicAction

if TYPE_CHECKING:
    from typing import Any

    from funtracks.data_model import Tracks
    from funtracks.data_model.tracks import Node


class UpdateNodeAttrs(BasicAction):
    """Action for user updates to node attributes. Cannot update protected
    attributes (time, area, track id), as these are controlled by internal application
    logic."""

    def __init__(
        self,
        tracks: Tracks,
        node: Node,
        attrs: dict[str, Any],
    ):
        """
        Args:
            tracks (Tracks): The tracks to update the node attributes for
            node (Node): The node to update the attributes for
            attrs (dict[str, Any]): A mapping from attribute name to list of new attribute
                values for the given nodes.

        Raises:
            ValueError: If a protected attribute is in the given attribute mapping.
        """
        super().__init__(tracks)
        # Cannot modify annotator-managed features or time
        protected_attrs = set(tracks.annotators.all_features.keys())
        protected_attrs.add(tracks.features.time_key)

        for attr in attrs:
            if attr in protected_attrs:
                raise ValueError(f"Cannot update attribute {attr} manually")
        self.node = node
        self.prev_attrs = {attr: self.tracks.get_node_attr(node, attr) for attr in attrs}
        self.new_attrs = attrs
        self._apply()

    def inverse(self) -> BasicAction:
        """Restore previous attributes"""
        return UpdateNodeAttrs(
            self.tracks,
            self.node,
            self.prev_attrs,
        )

    def _apply(self) -> None:
        """Set new attributes"""
        for attr, value in self.new_attrs.items():
            self.tracks._set_node_attr(self.node, attr, value)



import pytest

from funtracks.actions import (
    UpdateNodeAttrs,
)


@pytest.mark.parametrize("ndim", [3, 4])
def test_update_node_attrs(get_tracks, ndim):
    tracks = get_tracks(ndim=ndim, with_seg=True, is_solution=True)
    node = 1
    new_attr = {"score": 1.0}

    action = UpdateNodeAttrs(tracks, node, new_attr)
    assert tracks.get_node_attr(node, "score") == 1.0

    inverse = action.inverse()
    assert tracks.get_node_attr(node, "score") is None

    inverse.inverse()
    assert tracks.get_node_attr(node, "score") == 1.0


@pytest.mark.parametrize("attr", ["t", "area", "track_id"])
def test_update_protected_attr(get_tracks, attr):
    tracks = get_tracks(ndim=3, with_seg=True, is_solution=True)
    with pytest.raises(ValueError, match="Cannot update attribute .* manually"):
        UpdateNodeAttrs(tracks, 1, {attr: 2})



from __future__ import annotations

from typing import TYPE_CHECKING

from ._feature import Feature

if TYPE_CHECKING:
    pass


def Area(ndim: int = 3) -> Feature:
    """A regionprops feature for computing area or volume.

    Args:
        ndim (int): The number of dimensions of the tracks. Controls the display
            name.

    Returns:
        Feature: A feature dict representing area/volume
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": 1,
        "display_name": "Area" if ndim == 3 else "Volume",
        "required": True,
        "default_value": None,
    }


def Intensity() -> Feature:
    """A regionprops feature for computing the intensity.

    Returns:
        Feature: A feature dict representing intensity
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": 1,
        "display_name": "Intensity",
        "required": True,
        "default_value": None,
    }


def EllipsoidAxes(ndim: int = 3) -> Feature:
    """A regionprops feature for computing the ellipsoid axis radii.

    Args:
        ndim (int): The number of dimensions of the tracks. Controls the display
            name.

    Returns:
        Feature: A feature dict representing ellipsoid axes
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": 1,
        "display_name": "Ellipse axis radii" if ndim == 3 else "Ellipsoid axis radii",
        "required": True,
        "default_value": None,
    }


def Circularity(ndim: int = 3) -> Feature:
    """A regionprops feature for computing the circularity or sphericity.

    Args:
        ndim (int): The number of dimensions of the tracks. Controls the display
            name.

    Returns:
        Feature: A feature dict representing circularity/sphericity
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": 1,
        "display_name": "Circularity" if ndim == 3 else "Sphericity",
        "required": True,
        "default_value": None,
    }


def Perimeter(ndim: int = 3) -> Feature:
    """A regionprops feature for computing perimeter or surface area.

    Args:
        ndim (int): The number of dimensions of the tracks. Controls the display
            name.

    Returns:
        Feature: A feature dict representing perimeter/surface area
    """
    return {
        "feature_type": "node",
        "value_type": "float",
        "num_values": 1,
        "display_name": "Perimeter" if ndim == 3 else "Surface Area",
        "required": True,
        "default_value": None,
    }



import pytest

from funtracks.features import FeatureDict, Position, Time


class TestFeatureDict:
    def test_init(self):
        """Test basic initialization of FeatureDict"""
        features = {"time": Time(), "pos": Position(("y", "x"))}
        fd = FeatureDict(
            features=features, time_key="time", position_key="pos", tracklet_key=None
        )

        assert len(fd) == 2
        assert fd.time_key == "time"
        assert fd.position_key == "pos"
        assert fd["time"] == Time()
        assert fd["pos"] == Position(("y", "x"))

    def test_init_with_list_position(self):
        """Test initialization with list of position keys"""
        features = {
            "time": Time(),
            "y": {
                "feature_type": "node",
                "value_type": "float",
                "num_values": 1,
                "display_name": None,
                "recompute": False,
                "required": True,
                "default_value": None,
            },
            "x": {
                "feature_type": "node",
                "value_type": "float",
                "num_values": 1,
                "display_name": None,
                "recompute": False,
                "required": True,
                "default_value": None,
            },
        }
        fd = FeatureDict(
            features=features, time_key="time", position_key=["y", "x"], tracklet_key=None
        )

        assert len(fd) == 3
        assert fd.time_key == "time"
        assert fd.position_key == ["y", "x"]
        # Check that both position features exist
        assert "y" in fd
        assert "x" in fd

    def test_init_validation(self):
        """Test that init validates time and position keys exist"""
        features = {"time": Time(), "pos": Position(("y", "x"))}

        # Missing time key
        with pytest.raises(KeyError, match="time_key 'invalid' not found"):
            FeatureDict(
                features, time_key="invalid", position_key="pos", tracklet_key=None
            )

        # Missing position key
        with pytest.raises(KeyError, match="position_key 'invalid' not found"):
            FeatureDict(
                features, time_key="time", position_key="invalid", tracklet_key=None
            )

        # Missing one of multiple position keys
        with pytest.raises(KeyError, match="position_key 'z' not found"):
            FeatureDict(
                features, time_key="time", position_key=["z", "y"], tracklet_key=None
            )

    def test_node_features(self):
        """Test node_features property filters correctly"""
        features = {
            "time": Time(),
            "pos": Position(("y", "x")),
            "iou": {
                "feature_type": "edge",
                "value_type": "float",
                "num_values": 1,
                "display_name": "IoU",
                "recompute": True,
                "required": False,
                "default_value": None,
            },
        }
        fd = FeatureDict(features, time_key="time", position_key="pos", tracklet_key=None)

        node_feats = fd.node_features
        assert len(node_feats) == 2
        assert "time" in node_feats
        assert "pos" in node_feats
        assert "iou" not in node_feats

    def test_edge_features(self):
        """Test edge_features property filters correctly"""
        features = {
            "time": Time(),
            "pos": Position(("y", "x")),
            "iou": {
                "feature_type": "edge",
                "value_type": "float",
                "num_values": 1,
                "display_name": "IoU",
                "recompute": True,
                "required": False,
                "default_value": None,
            },
        }
        fd = FeatureDict(features, time_key="time", position_key="pos", tracklet_key=None)

        edge_feats = fd.edge_features
        assert len(edge_feats) == 1
        assert "iou" in edge_feats
        assert "time" not in edge_feats

    @pytest.mark.parametrize("composite_position", [True, False])
    def test_json_dump_and_load(self, composite_position):
        """Test JSON serialization and deserialization"""
        if composite_position:
            pos_key = "pos"
            features = {"time": Time(), pos_key: Position(("y", "x"))}
        else:
            pos_key = ["y", "x"]
            features = {
                "time": Time(),
                "y": {
                    "feature_type": "node",
                    "value_type": "float",
                    "num_values": 1,
                    "display_name": None,
                    "recompute": False,
                    "required": True,
                    "default_value": None,
                },
                "x": {
                    "feature_type": "node",
                    "value_type": "float",
                    "num_values": 1,
                    "display_name": None,
                    "recompute": False,
                    "required": True,
                    "default_value": None,
                },
            }

        fd = FeatureDict(
            features, time_key="time", position_key=pos_key, tracklet_key=None
        )
        json_dict = fd.dump_json()

        assert "FeatureDict" in json_dict
        assert "features" in json_dict["FeatureDict"]
        assert "time_key" in json_dict["FeatureDict"]
        assert "position_key" in json_dict["FeatureDict"]
        assert "tracklet_key" in json_dict["FeatureDict"]

        # Load back from JSON
        loaded_fd = FeatureDict.from_json(json_dict)
        assert loaded_fd.time_key == fd.time_key
        assert loaded_fd.position_key == fd.position_key
        assert len(loaded_fd) == len(fd)

        # Check features match
        for key in fd:
            assert key in loaded_fd
            assert loaded_fd[key] == fd[key]

    def test_dict_behavior(self):
        """Test that FeatureDict behaves like a dict"""
        features = {"time": Time(), "pos": Position(("y", "x"))}
        fd = FeatureDict(features, time_key="time", position_key="pos", tracklet_key=None)

        # Can iterate over keys
        keys = list(fd.keys())
        assert "time" in keys
        assert "pos" in keys

        # Can access like a dict
        assert fd["time"] == Time()
        assert fd["pos"] == Position(("y", "x"))

        # Can check membership
        assert "time" in fd
        assert "nonexistent" not in fd
